# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the photometry module.
"""

from contextlib import nullcontext

import astropy
import numpy as np
import pytest
from astropy.convolution.utils import discretize_model
from astropy.modeling import Fittable2DModel, Parameter
from astropy.modeling.fitting import LevMarLSQFitter, SimplexLSQFitter
from astropy.modeling.models import Gaussian2D, Moffat2D
from astropy.stats import SigmaClip, gaussian_sigma_to_fwhm
from astropy.table import Table
from astropy.utils import minversion
from astropy.utils.exceptions import (AstropyDeprecationWarning,
                                      AstropyUserWarning)
from numpy.testing import assert_allclose, assert_array_equal, assert_equal

from photutils.background import MMMBackground, StdBackgroundRMS
from photutils.datasets import (make_gaussian_prf_sources_image,
                                make_noise_image)
from photutils.detection import DAOStarFinder, IRAFStarFinder
from photutils.psf.groupstars import DAOGroup
from photutils.psf.models import FittableImageModel, IntegratedGaussianPRF
from photutils.psf.photometry_depr import (BasicPSFPhotometry,
                                           DAOPhotPSFPhotometry,
                                           IterativelySubtractedPSFPhotometry)
from photutils.psf.utils import prepare_psf_model
from photutils.tests.helper import PYTEST_LT_80
from photutils.utils._optional_deps import HAS_SCIPY


def make_psf_photometry_objs(std=1, sigma_psf=1):
    """
    Produces baseline photometry objects which are then
    modified as-needed in specific tests below
    """
    with pytest.warns(AstropyDeprecationWarning):
        daofind = DAOStarFinder(threshold=5.0 * std,
                                fwhm=sigma_psf * gaussian_sigma_to_fwhm)
        daogroup = DAOGroup(1.5 * sigma_psf * gaussian_sigma_to_fwhm)
        threshold = 5.0 * std
        fwhm = sigma_psf * gaussian_sigma_to_fwhm
        crit_separation = 1.5 * sigma_psf * gaussian_sigma_to_fwhm

        daofind = DAOStarFinder(threshold=threshold, fwhm=fwhm)
        daogroup = DAOGroup(crit_separation)
        mode_bkg = MMMBackground()
        psf_model = IntegratedGaussianPRF(sigma=sigma_psf)
        fitter = LevMarLSQFitter()

        basic_phot_obj = BasicPSFPhotometry(finder=daofind,
                                            group_maker=daogroup,
                                            bkg_estimator=mode_bkg,
                                            psf_model=psf_model,
                                            fitter=fitter,
                                            fitshape=(11, 11))

        iter_phot_obj = IterativelySubtractedPSFPhotometry(
            finder=daofind, group_maker=daogroup, bkg_estimator=mode_bkg,
            psf_model=psf_model, fitter=fitter, niters=1, fitshape=(11, 11))

        dao_phot_obj = DAOPhotPSFPhotometry(crit_separation=crit_separation,
                                            threshold=threshold, fwhm=fwhm,
                                            psf_model=psf_model,
                                            fitshape=(11, 11),
                                            niters=1)

        return (basic_phot_obj, iter_phot_obj, dao_phot_obj)


sigma_psfs = []

# A group of two overlapped stars and an isolated one
sigma_psfs.append(2)
sources1 = Table()
sources1['flux'] = [800, 1000, 1200]
sources1['x_0'] = [13, 18, 25]
sources1['y_0'] = [16, 16, 25]
sources1['sigma'] = [sigma_psfs[-1]] * 3
sources1['theta'] = [0] * 3
sources1['id'] = [1, 2, 3]
sources1['group_id'] = [1, 1, 2]


# one single group with four stars.
sigma_psfs.append(2)
sources2 = Table()
sources2['flux'] = [700, 800, 700, 800]
sources2['x_0'] = [12, 17, 12, 17]
sources2['y_0'] = [15, 15, 20, 20]
sources2['sigma'] = [sigma_psfs[-1]] * 4
sources2['theta'] = [0] * 4
sources2['id'] = [1, 2, 3, 4]
sources2['group_id'] = [1, 1, 1, 1]

# one faint star and one brither companion
# although they are in the same group, the detection algorithm
# is not able to detect the fainter star, hence photometry should
# be performed with niters > 1 or niters=None
sigma_psfs.append(2)
sources3 = Table()
sources3['flux'] = [10000, 1000]
sources3['x_0'] = [18, 13]
sources3['y_0'] = [17, 19]
sources3['sigma'] = [sigma_psfs[-1]] * 2
sources3['theta'] = [0] * 2
sources3['id'] = [1] * 2
sources3['group_id'] = [1] * 2
sources3['iter_detected'] = [1, 2]


@pytest.mark.skipif(not HAS_SCIPY, reason='scipy is required')
@pytest.mark.parametrize('sigma_psf, sources', [(sigma_psfs[2], sources3)])
def test_psf_photometry_niters(sigma_psf, sources):
    img_shape = (32, 32)
    # generate image with read-out noise (Gaussian) and
    # background noise (Poisson)
    image = (make_gaussian_prf_sources_image(img_shape, sources)
             + make_noise_image(img_shape, distribution='poisson', mean=6.0,
                                seed=0)
             + make_noise_image(img_shape, distribution='gaussian', mean=0.0,
                                stddev=2.0, seed=0))
    cp_image = image.copy()
    sigma_clip = SigmaClip(sigma=3.0)
    bkgrms = StdBackgroundRMS(sigma_clip)
    std = bkgrms(image)

    phot_obj = make_psf_photometry_objs(std, sigma_psf)[1:3]
    for iter_phot_obj in phot_obj:
        iter_phot_obj.niters = None

        with pytest.warns(AstropyDeprecationWarning):
            result_tab = iter_phot_obj(image)
            residual_image = iter_phot_obj.get_residual_image()

        assert (result_tab['x_0_unc'] < 1.96 * sigma_psf
                / np.sqrt(sources['flux'])).all()
        assert (result_tab['y_0_unc'] < 1.96 * sigma_psf
                / np.sqrt(sources['flux'])).all()
        assert (result_tab['flux_unc'] < 1.96
                * np.sqrt(sources['flux'])).all()

        assert_allclose(result_tab['x_fit'], sources['x_0'], rtol=1e-1)
        assert_allclose(result_tab['y_fit'], sources['y_0'], rtol=1e-1)
        assert_allclose(result_tab['flux_fit'], sources['flux'], rtol=1e-1)
        assert_array_equal(result_tab['id'], sources['id'])
        assert_array_equal(result_tab['group_id'], sources['group_id'])
        assert_array_equal(result_tab['iter_detected'],
                           sources['iter_detected'])
        assert_allclose(np.mean(residual_image), 0.0, atol=1e1)

        # make sure image is note overwritten
        assert_array_equal(cp_image, image)


@pytest.mark.filterwarnings('ignore:Both init_guesses and finder are '
                            'different than None')
@pytest.mark.filterwarnings('ignore:No sources were found')
@pytest.mark.skipif(not HAS_SCIPY, reason='scipy is required')
@pytest.mark.parametrize('sigma_psf, sources',
                         [(sigma_psfs[0], sources1),
                          (sigma_psfs[1], sources2)])
def test_psf_photometry_oneiter(sigma_psf, sources):
    """
    Tests in an image with a group of two overlapped stars and an
    isolated one.
    """

    img_shape = (32, 32)
    # generate image with read-out noise (Gaussian) and
    # background noise (Poisson)
    image = (make_gaussian_prf_sources_image(img_shape, sources)
             + make_noise_image(img_shape, distribution='poisson', mean=6.0,
                                seed=0)
             + make_noise_image(img_shape, distribution='gaussian', mean=0.0,
                                stddev=2.0, seed=0))
    cp_image = image.copy()

    sigma_clip = SigmaClip(sigma=3.0)
    bkgrms = StdBackgroundRMS(sigma_clip)
    std = bkgrms(image)
    phot_objs = make_psf_photometry_objs(std, sigma_psf)

    for phot_proc in phot_objs:
        with pytest.warns(AstropyDeprecationWarning):
            result_tab = phot_proc(image)
            residual_image = phot_proc.get_residual_image()
        assert (result_tab['x_0_unc'] < 1.96 * sigma_psf
                / np.sqrt(sources['flux'])).all()
        assert (result_tab['y_0_unc'] < 1.96 * sigma_psf
                / np.sqrt(sources['flux'])).all()
        assert (result_tab['flux_unc'] < 1.96
                * np.sqrt(sources['flux'])).all()
        assert_allclose(result_tab['x_fit'], sources['x_0'], rtol=1e-1)
        assert_allclose(result_tab['y_fit'], sources['y_0'], rtol=1e-1)
        assert_allclose(result_tab['flux_fit'], sources['flux'], rtol=1e-1)
        assert_array_equal(result_tab['id'], sources['id'])
        assert_array_equal(result_tab['group_id'], sources['group_id'])
        assert_allclose(np.mean(residual_image), 0.0, atol=1e1)

        # test fixed photometry
        phot_proc.psf_model.x_0.fixed = True
        phot_proc.psf_model.y_0.fixed = True

        pos = Table(names=['x_0', 'y_0'], data=[sources['x_0'],
                                                sources['y_0']])
        cp_pos = pos.copy()

        with pytest.warns(AstropyDeprecationWarning):
            result_tab = phot_proc(image, init_guesses=pos)
            residual_image = phot_proc.get_residual_image()
        assert 'x_0_unc' not in result_tab.colnames
        assert 'y_0_unc' not in result_tab.colnames
        assert (result_tab['flux_unc'] < 1.96
                * np.sqrt(sources['flux'])).all()
        assert_array_equal(result_tab['x_fit'], sources['x_0'])
        assert_array_equal(result_tab['y_fit'], sources['y_0'])
        assert_allclose(result_tab['flux_fit'], sources['flux'], rtol=1e-1)
        assert_array_equal(result_tab['id'], sources['id'])
        assert_array_equal(result_tab['group_id'], sources['group_id'])
        assert_allclose(np.mean(residual_image), 0.0, atol=1e1)

        # make sure image is not overwritten
        assert_array_equal(cp_image, image)

        # make sure initial guess table is not modified
        assert_array_equal(cp_pos, pos)

        # resets fixed positions
        phot_proc.psf_model.x_0.fixed = False
        phot_proc.psf_model.y_0.fixed = False


@pytest.mark.skipif(not HAS_SCIPY, reason='scipy is required')
def test_niters_errors():
    iter_phot_obj = make_psf_photometry_objs()[1]

    # tests that niters is set to an integer even if the user inputs
    # a float
    iter_phot_obj.niters = 1.1
    assert_equal(iter_phot_obj.niters, 1)

    # test that a ValueError is raised if niters <= 0
    with pytest.raises(ValueError):
        iter_phot_obj.niters = 0

    # test that it's OK to set niters to None
    iter_phot_obj.niters = None


@pytest.mark.skipif(not HAS_SCIPY, reason='scipy is required')
def test_fitshape_errors():
    basic_phot_obj = make_psf_photometry_objs()[0]

    # first make sure setting to a scalar does the right thing (and makes
    # no errors)
    basic_phot_obj.fitshape = 11
    assert np.all(basic_phot_obj.fitshape == (11, 11))

    # test that a ValuError is raised if fitshape has even components
    with pytest.raises(ValueError):
        basic_phot_obj.fitshape = (2, 2)
    with pytest.raises(ValueError):
        basic_phot_obj.fitshape = 2

    # test that a ValueError is raised if fitshape has non positive
    # components
    with pytest.raises(ValueError):
        basic_phot_obj.fitshape = (-1, 0)

    # test that a ValueError is raised if fitshape has more than two
    # dimensions
    with pytest.raises(ValueError):
        basic_phot_obj.fitshape = (3, 3, 3)


@pytest.mark.skipif(not HAS_SCIPY, reason='scipy is required')
def test_aperture_radius_errors():
    basic_phot_obj = make_psf_photometry_objs()[0]

    # test that aperture_radius was set to None by default
    assert_equal(basic_phot_obj.aperture_radius, None)

    # test that a ValueError is raised if aperture_radius is non positive
    with pytest.raises(ValueError):
        basic_phot_obj.aperture_radius = -3


@pytest.mark.skipif(not HAS_SCIPY, reason='scipy is required')
def test_finder_errors():
    with pytest.warns(AstropyDeprecationWarning):
        iter_phot_obj = make_psf_photometry_objs()[1]

        with pytest.raises(ValueError):
            iter_phot_obj.finder = None

        with pytest.raises(ValueError):
            iter_phot_obj = IterativelySubtractedPSFPhotometry(
                finder=None, group_maker=DAOGroup(1),
                bkg_estimator=MMMBackground(),
                psf_model=IntegratedGaussianPRF(1), fitshape=(11, 11))


@pytest.mark.skipif(not HAS_SCIPY, reason='scipy is required')
def test_finder_positions_warning():
    basic_phot_obj = make_psf_photometry_objs(sigma_psf=2)[0]
    positions = Table()
    positions['x_0'] = [12.8, 18.2, 25.3]
    positions['y_0'] = [15.7, 16.5, 25.1]

    image = (make_gaussian_prf_sources_image((32, 32), sources1)
             + make_noise_image((32, 32), distribution='poisson', mean=6.0,
                                seed=0))

    match = 'Both init_guesses and finder are different than None'
    ctx1 = pytest.warns(AstropyUserWarning, match=match)
    if PYTEST_LT_80:
        ctx2 = nullcontext()
    else:
        ctx2 = pytest.warns(AstropyDeprecationWarning)
    with ctx1, ctx2:
        result_tab = basic_phot_obj(image=image, init_guesses=positions)
        assert_array_equal(result_tab['x_0'], positions['x_0'])
        assert_array_equal(result_tab['y_0'], positions['y_0'])
        assert_allclose(result_tab['x_fit'], positions['x_0'], rtol=1e-1)
        assert_allclose(result_tab['y_fit'], positions['y_0'], rtol=1e-1)

    basic_phot_obj.finder = None
    with pytest.raises(ValueError):
        result_tab = basic_phot_obj(image=image)


@pytest.mark.filterwarnings('ignore:The fit may be unsuccessful')
@pytest.mark.skipif(not HAS_SCIPY, reason='scipy is required')
def test_aperture_radius():
    img_shape = (32, 32)

    # generate image with read-out noise (Gaussian) and
    # background noise (Poisson)
    image = (make_gaussian_prf_sources_image(img_shape, sources1)
             + make_noise_image(img_shape, distribution='poisson', mean=6.0,
                                seed=0)
             + make_noise_image(img_shape, distribution='gaussian', mean=0.0,
                                stddev=2.0, seed=0))

    basic_phot_obj = make_psf_photometry_objs()[0]

    # test that aperture radius is properly set whenever the PSF model has
    # a `fwhm` attribute
    class PSFModelWithFWHM(Fittable2DModel):
        x_0 = Parameter(default=1)
        y_0 = Parameter(default=1)
        flux = Parameter(default=1)
        fwhm = Parameter(default=5)

        def __init__(self, fwhm=fwhm.default):
            super().__init__(fwhm=fwhm)

        def evaluate(self, x, y, x_0, y_0, flux, fwhm):
            return flux / (fwhm * (x - x_0)**2 * (y - y_0)**2)

    psf_model = PSFModelWithFWHM()
    basic_phot_obj.psf_model = psf_model
    with pytest.warns(AstropyDeprecationWarning):
        basic_phot_obj(image)
    assert_equal(basic_phot_obj.aperture_radius, psf_model.fwhm.value)


PARS_TO_SET_0 = {'x_0': 'x_0', 'y_0': 'y_0', 'flux_0': 'flux'}
PARS_TO_OUTPUT_0 = {'x_fit': 'x_0', 'y_fit': 'y_0', 'flux_fit': 'flux'}
PARS_TO_SET_1 = PARS_TO_SET_0.copy()
PARS_TO_SET_1['sigma_0'] = 'sigma'
PARS_TO_OUTPUT_1 = PARS_TO_OUTPUT_0.copy()
PARS_TO_OUTPUT_1['sigma_fit'] = 'sigma'


@pytest.mark.parametrize('actual_pars_to_set, actual_pars_to_output,'
                         'is_sigma_fixed', [(PARS_TO_SET_0, PARS_TO_OUTPUT_0,
                                             True),
                                            (PARS_TO_SET_1, PARS_TO_OUTPUT_1,
                                             False)])
@pytest.mark.skipif(not HAS_SCIPY, reason='scipy is required')
def test_define_fit_param_names(actual_pars_to_set, actual_pars_to_output,
                                is_sigma_fixed):
    psf_model = IntegratedGaussianPRF()
    psf_model.sigma.fixed = is_sigma_fixed

    basic_phot_obj = make_psf_photometry_objs()[0]
    basic_phot_obj.psf_model = psf_model

    with pytest.warns(AstropyDeprecationWarning):
        basic_phot_obj._define_fit_param_names()
    assert_equal(basic_phot_obj._pars_to_set, actual_pars_to_set)
    assert_equal(basic_phot_obj._pars_to_output, actual_pars_to_output)


# tests previously written to psf_photometry

PSF_SIZE = 11
GAUSSIAN_WIDTH = 1.0
IMAGE_SIZE = 101

# Position and FLUXES of test sources
INTAB = Table([[50.0, 23, 12, 86], [50.0, 83, 80, 84],
               [np.pi * 10, 3.654, 20.0, 80 / np.sqrt(3)]],
              names=['x_0', 'y_0', 'flux_0'])

# Create test psf
psf_model = Gaussian2D(1.0 / (2 * np.pi * GAUSSIAN_WIDTH ** 2), PSF_SIZE // 2,
                       PSF_SIZE // 2, GAUSSIAN_WIDTH, GAUSSIAN_WIDTH)
test_psf = discretize_model(psf_model, (0, PSF_SIZE), (0, PSF_SIZE),
                            mode='oversample')

# Set up grid for test image
image = np.zeros((IMAGE_SIZE, IMAGE_SIZE))

# Add sources to test image
for x, y, flux in INTAB:
    model = Gaussian2D(flux / (2 * np.pi * GAUSSIAN_WIDTH ** 2),
                       x, y, GAUSSIAN_WIDTH, GAUSSIAN_WIDTH)
    image += discretize_model(model, (0, IMAGE_SIZE), (0, IMAGE_SIZE),
                              mode='oversample')

# Some tests require an image with wider sources.
WIDE_GAUSSIAN_WIDTH = 3.0
WIDE_INTAB = Table([[50, 23.2], [50.5, 1], [10, 20]],
                   names=['x_0', 'y_0', 'flux_0'])
wide_image = np.zeros((IMAGE_SIZE, IMAGE_SIZE))

# Add sources to test image
for x, y, flux in WIDE_INTAB:
    model = Gaussian2D(flux / (2 * np.pi * WIDE_GAUSSIAN_WIDTH ** 2),
                       x, y, WIDE_GAUSSIAN_WIDTH, WIDE_GAUSSIAN_WIDTH)
    wide_image += discretize_model(model, (0, IMAGE_SIZE), (0, IMAGE_SIZE),
                                   mode='oversample')


@pytest.mark.skipif(not HAS_SCIPY, reason='scipy is required')
def test_default_aperture_radius():
    """
    Test psf_photometry with non-Gaussian model, such that it raises a
    warning about aperture_radius.
    """

    def tophatfinder(image, mask=None):
        """Simple top hat finder function for use with a top hat PRF."""
        fluxes = np.unique(image[image > 1])
        table = Table(names=['id', 'xcentroid', 'ycentroid', 'flux'],
                      dtype=[int, float, float, float])
        for n, f in enumerate(fluxes):
            ys, xs = np.where(image == f)
            x = np.mean(xs)
            y = np.mean(ys)
            table.add_row([int(n + 1), x, y, f * 9])
        table.sort(['flux'])

        return table

    with pytest.warns(AstropyDeprecationWarning):
        prf = np.zeros((7, 7), float)
        prf[2:5, 2:5] = 1 / 9
        prf = FittableImageModel(prf)

        img = np.zeros((50, 50), float)
        x0 = [38, 20, 35]
        y0 = [20, 5, 40]
        f0 = [50, 100, 200]
        for x, y, f in zip(x0, y0, f0):
            img[y - 1:y + 2, x - 1:x + 2] = f / 9

        intab = Table(data=[[37, 19.6, 34.9], [19.6, 4.5, 40.1]],
                      names=['x_0', 'y_0'])

        basic_phot = BasicPSFPhotometry(group_maker=DAOGroup(2),
                                        bkg_estimator=None, psf_model=prf,
                                        fitshape=7, finder=tophatfinder)
        # Test for init_guesses is None
        match = 'aperture_radius is None and could not be determined'
        with pytest.warns(AstropyUserWarning, match=match):
            results = basic_phot(image=img)
        assert_allclose(results['flux_fit'], f0, rtol=0.05)

        # Have to reset the object or it saves any updates, and we wish to
        # re-verify the aperture_radius assignment
        basic_phot = BasicPSFPhotometry(group_maker=DAOGroup(2),
                                        bkg_estimator=None, psf_model=prf,
                                        fitshape=7)

        # Test for init_guesses is not None, but lacks a flux_0 column
        match = 'aperture_radius is None and could not be determined'
        with pytest.warns(AstropyUserWarning, match=match):
            results = basic_phot(image=img, init_guesses=intab)
        assert_allclose(results['flux_fit'], f0, rtol=0.05)

        iter_phot = IterativelySubtractedPSFPhotometry(finder=tophatfinder,
                                                       group_maker=DAOGroup(2),
                                                       bkg_estimator=None,
                                                       psf_model=prf,
                                                       fitshape=7, niters=2)

        match1 = 'aperture_radius is None and could not be determined'
        ctx1 = pytest.warns(AstropyUserWarning, match=match1)
        if PYTEST_LT_80:
            ctx2 = nullcontext()
        else:
            match2 = 'Both init_guesses and finder are different than None'
            ctx2 = pytest.warns(AstropyUserWarning, match=match2)
        with ctx1, ctx2:
            results = iter_phot(image=img, init_guesses=intab)
            assert_allclose(results['flux_fit'], f0, rtol=0.05)

        iter_phot = IterativelySubtractedPSFPhotometry(
            finder=tophatfinder, group_maker=DAOGroup(2), bkg_estimator=None,
            psf_model=prf, fitshape=7, niters=2)

        # Test for init_guesses is None
        match = 'aperture_radius is None and could not be determined'
        with pytest.warns(AstropyUserWarning, match=match):
            results = iter_phot(image=img)
        assert_allclose(results['flux_fit'], f0, rtol=0.05)


@pytest.mark.skipif(not HAS_SCIPY, reason='scipy is required')
def test_psf_boundary_gaussian():
    """
    Test psf_photometry with discrete PRF model at the boundary of the data.
    """
    with pytest.warns(AstropyDeprecationWarning):
        psf = IntegratedGaussianPRF(GAUSSIAN_WIDTH)

        basic_phot = BasicPSFPhotometry(group_maker=DAOGroup(2),
                                        bkg_estimator=None, psf_model=psf,
                                        fitshape=7)

        intab = Table(data=[[1], [1]], names=['x_0', 'y_0'])
        f = basic_phot(image=image, init_guesses=intab)
        assert_allclose(f['flux_fit'], 0, atol=1e-8)


@pytest.mark.skipif(not HAS_SCIPY, reason='scipy is required')
def test_psf_photometry_gaussian():
    """
    Test psf_photometry with Gaussian PSF model.
    """
    with pytest.warns(AstropyDeprecationWarning):
        psf = IntegratedGaussianPRF(sigma=GAUSSIAN_WIDTH)

        basic_phot = BasicPSFPhotometry(group_maker=DAOGroup(2),
                                        bkg_estimator=None, psf_model=psf,
                                        fitshape=7)
        f = basic_phot(image=image, init_guesses=INTAB)
        for n in ['x', 'y', 'flux']:
            assert_allclose(f[n + '_0'], f[n + '_fit'], rtol=1e-3)


@pytest.mark.skipif(not HAS_SCIPY, reason='scipy is required')
@pytest.mark.parametrize('renormalize_psf', (True, False))
def test_psf_photometry_gaussian2(renormalize_psf):
    """
    Test psf_photometry with Gaussian PSF model from Astropy.
    """
    with pytest.warns(AstropyDeprecationWarning):
        psf = Gaussian2D(1.0 / (2 * np.pi * GAUSSIAN_WIDTH ** 2),
                         PSF_SIZE // 2, PSF_SIZE // 2, GAUSSIAN_WIDTH,
                         GAUSSIAN_WIDTH)
        psf = prepare_psf_model(psf, xname='x_mean', yname='y_mean',
                                renormalize_psf=renormalize_psf)

        basic_phot = BasicPSFPhotometry(group_maker=DAOGroup(2),
                                        bkg_estimator=None, psf_model=psf,
                                        fitshape=7)
        match = ('aperture_radius is None and could not be determined by '
                 'psf_model')
        with pytest.warns(AstropyUserWarning, match=match):
            f = basic_phot(image=image, init_guesses=INTAB)

        for n in ['x', 'y']:
            assert_allclose(f[n + '_0'], f[n + '_fit'], rtol=1e-1)
        assert_allclose(f['flux_0'], f['flux_fit'], rtol=1e-1)


@pytest.mark.skipif(not HAS_SCIPY, reason='scipy is required')
def test_psf_photometry_moffat():
    """
    Test psf_photometry with Moffat PSF model from Astropy.
    """
    with pytest.warns(AstropyDeprecationWarning):
        psf = Moffat2D(1.0 / (2 * np.pi * GAUSSIAN_WIDTH ** 2), PSF_SIZE // 2,
                       PSF_SIZE // 2, 1, 1)
        psf = prepare_psf_model(psf, xname='x_0', yname='y_0',
                                renormalize_psf=False)

        basic_phot = BasicPSFPhotometry(group_maker=DAOGroup(2),
                                        bkg_estimator=None, psf_model=psf,
                                        fitshape=7)
        match = ('aperture_radius is None and could not be determined by '
                 'psf_model')
        with pytest.warns(AstropyUserWarning, match=match):
            f = basic_phot(image=image, init_guesses=INTAB)
        f.pprint(max_width=-1)

        for n in ['x', 'y']:
            assert_allclose(f[n + '_0'], f[n + '_fit'], rtol=1e-3)
        # image was created with a gaussian, so flux won't match exactly
        assert_allclose(f['flux_0'], f['flux_fit'], rtol=1e-1)


@pytest.mark.skipif(not HAS_SCIPY, reason='scipy is required')
def test_psf_fitting_data_on_edge():
    """
    No mask is input explicitly here, but source 2 is so close to the
    edge that the subarray that's extracted gets a mask internally.
    """
    with pytest.warns(AstropyDeprecationWarning):
        psf_guess = IntegratedGaussianPRF(flux=1, sigma=WIDE_GAUSSIAN_WIDTH)
        psf_guess.flux.fixed = psf_guess.x_0.fixed = False
        psf_guess.y_0.fixed = False
        basic_phot = BasicPSFPhotometry(group_maker=DAOGroup(2),
                                        bkg_estimator=None,
                                        psf_model=psf_guess, fitshape=7)

        outtab = basic_phot(image=wide_image, init_guesses=WIDE_INTAB)

        for n in ['x', 'y', 'flux']:
            assert_allclose(outtab[n + '_0'], outtab[n + '_fit'],
                            rtol=0.05, atol=0.1)


@pytest.mark.filterwarnings('ignore:Both init_guesses and finder '
                            'are different than None')
@pytest.mark.skipif(not HAS_SCIPY, reason='scipy is required')
@pytest.mark.parametrize('sigma_psf, sources', [(sigma_psfs[2], sources3)])
def test_psf_extra_output_cols(sigma_psf, sources):
    """
    Test the handling of a non-None extra_output_cols
    """
    with pytest.warns(AstropyDeprecationWarning):
        psf_model = IntegratedGaussianPRF(sigma=sigma_psf)
        tshape = (32, 32)
        image = (make_gaussian_prf_sources_image(tshape, sources)
                 + make_noise_image(tshape, distribution='poisson', mean=6.0,
                                    seed=0)
                 + make_noise_image(tshape, distribution='gaussian', mean=0.0,
                                    stddev=2.0, seed=0))

        init_guess1 = None
        init_guess2 = Table(names=['x_0', 'y_0', 'sharpness', 'roundness1',
                                   'roundness2'],
                            data=[[17.4], [16], [0.4], [0], [0]])
        init_guess3 = Table(names=['x_0', 'y_0'],
                            data=[[17.4], [16]])
        init_guess4 = Table(names=['x_0', 'y_0', 'sharpness'],
                            data=[[17.4], [16], [0.4]])
        for i, init_guesses in enumerate([init_guess1, init_guess2,
                                          init_guess3, init_guess4]):
            dao_phot = DAOPhotPSFPhotometry(crit_separation=8, threshold=40,
                                            fwhm=4 * np.sqrt(2 * np.log(2)),
                                            psf_model=psf_model,
                                            fitshape=(11, 11),
                                            extra_output_cols=['sharpness',
                                                               'roundness1',
                                                               'roundness2'])
            phot_results = dao_phot(image, init_guesses=init_guesses)
            # test that the original required columns are also passed
            # back, as well as extra_output_cols
            assert np.all([name in phot_results.colnames for name in
                           ['x_0', 'y_0']])
            assert np.all([name in phot_results.colnames for name in
                           ['sharpness', 'roundness1', 'roundness2']])
            assert len(phot_results) == 2
            # checks to verify that half-passing init_guesses results
            # in NaN output for extra_output_cols not passed as initial
            # guesses
            if i == 2:  # init_guess3
                assert np.all(np.all(np.isnan(phot_results[o])) for o in
                              ['sharpness', 'roundness1', 'roundness2'])
            if i == 3:  # init_guess4
                assert np.all(np.all(np.isnan(phot_results[o])) for o in
                              ['roundness1', 'roundness2'])
                assert np.all(~np.isnan(phot_results['sharpness']))


@pytest.fixture(params=[2, 3])
def overlap_image(request):
    if request.param == 2:
        close_tab = Table([[50.0, 53.0], [50.0, 50.0], [25.0, 25.0]],
                          names=['x_0', 'y_0', 'flux_0'])
    elif request.param == 3:
        close_tab = Table([[50.0, 55.0, 50.0], [50.0, 50.0, 55.0],
                           [25.0, 25.0, 25.0]], names=['x_0', 'y_0', 'flux_0'])
    else:
        raise ValueError

    # Add sources to test image
    close_image = np.zeros((IMAGE_SIZE, IMAGE_SIZE))
    for x, y, flux in close_tab:
        close_model = Gaussian2D(flux / (2 * np.pi * GAUSSIAN_WIDTH ** 2),
                                 x, y, GAUSSIAN_WIDTH, GAUSSIAN_WIDTH)
        close_image += discretize_model(close_model, (0, IMAGE_SIZE),
                                        (0, IMAGE_SIZE), mode='oversample')
    return close_image


@pytest.mark.skipif(not HAS_SCIPY, reason='scipy is required')
def test_psf_fitting_group(overlap_image):
    """
    Test psf_photometry when two input stars are close and need to be
    fit together.
    """
    with pytest.warns(AstropyDeprecationWarning):
        from photutils.background import MADStdBackgroundRMS

        # There are a few models here that fail, be it something created
        # by EPSFBuilder or simpler the Moffat2D one unprepared_psf =
        # Moffat2D(amplitude=1, gamma=2, alpha=2.8, x_0=0, y_0=0) psf =
        # prepare_psf_model(unprepared_psf, xname='x_0', yname='y_0',
        # fluxname=None)
        psf = prepare_psf_model(Gaussian2D(), renormalize_psf=False)

        psf.fwhm = Parameter('fwhm', 'this is not the way to add this I think')
        psf.fwhm.value = 10

        separation_crit = 10

        # choose low threshold and fwhm to find stars no matter what
        basic_phot = BasicPSFPhotometry(finder=DAOStarFinder(1, 1),
                                        group_maker=DAOGroup(separation_crit),
                                        bkg_estimator=MADStdBackgroundRMS(),
                                        fitter=LevMarLSQFitter(),
                                        psf_model=psf,
                                        fitshape=31)
        # this should not raise AttributeError: Attribute "offset_0_0"
        # not found
        basic_phot(image=overlap_image)


@pytest.mark.skipif(not HAS_SCIPY, reason='scipy is required')
def test_finder_return_none():
    """
    Test psf_photometry with finder that does not return None if no
    sources are detected, to test Iterative PSF fitting.
    """

    def tophatfinder(image, mask=None):
        """Simple top hat finder function for use with a top hat PRF."""
        fluxes = np.unique(image[image > 1])
        table = Table(names=['id', 'xcentroid', 'ycentroid', 'flux'],
                      dtype=[int, float, float, float])
        for n, f in enumerate(fluxes):
            ys, xs = np.where(image == f)
            x = np.mean(xs)
            y = np.mean(ys)
            table.add_row([int(n + 1), x, y, f * 9])
        table.sort(['flux'])

        return table

    with pytest.warns(AstropyDeprecationWarning):
        prf = np.zeros((7, 7), float)
        prf[2:5, 2:5] = 1 / 9
        prf = FittableImageModel(prf)

        img = np.zeros((50, 50), float)
        x0 = [38, 20, 35]
        y0 = [20, 5, 40]
        f0 = [50, 100, 200]
        for x, y, f in zip(x0, y0, f0):
            img[y - 1:y + 2, x - 1:x + 2] = f / 9

        intab = Table(data=[[37, 19.6, 34.9], [19.6, 4.5, 40.1],
                            [45, 103, 210]], names=['x_0', 'y_0', 'flux_0'])

        iter_phot = IterativelySubtractedPSFPhotometry(finder=tophatfinder,
                                                       group_maker=DAOGroup(2),
                                                       bkg_estimator=None,
                                                       psf_model=prf,
                                                       fitshape=7, niters=2,
                                                       aperture_radius=3)

        match = 'Both init_guesses and finder are different than None'
        with pytest.warns(AstropyUserWarning, match=match):
            results = iter_phot(image=img, init_guesses=intab)
        assert_allclose(results['flux_fit'], f0, rtol=0.05)


@pytest.mark.skipif(not HAS_SCIPY, reason='scipy is required')
def test_psf_photometry_uncertainties():
    """
    Test an Astropy fitter that does not return a parameter
    covariance matrix (param_cov). The output table should not
    contain flux_unc, x_0_unc, and y_0_unc columns.
    """
    with pytest.warns(AstropyDeprecationWarning):
        psf = IntegratedGaussianPRF(sigma=GAUSSIAN_WIDTH)

        basic_phot = BasicPSFPhotometry(group_maker=DAOGroup(2),
                                        bkg_estimator=None, psf_model=psf,
                                        fitter=SimplexLSQFitter(),
                                        fitshape=7)
        match = 'The fit may be unsuccessful'
        with pytest.warns(AstropyUserWarning, match=match):
            phot_tbl = basic_phot(image=image, init_guesses=INTAB)
        columns = ('flux_unc', 'x_0_unc', 'y_0_unc')
        for column in columns:
            assert column not in phot_tbl.colnames


@pytest.mark.skipif(not HAS_SCIPY, reason='scipy is required')
def test_re_use_result_as_initial_guess():
    img_shape = (32, 32)
    # generate image with read-out noise (Gaussian) and
    # background noise (Poisson)
    image = (make_gaussian_prf_sources_image(img_shape, sources1)
             + make_noise_image(img_shape, distribution='poisson', mean=6.0,
                                seed=0)
             + make_noise_image(img_shape, distribution='gaussian', mean=0.0,
                                stddev=2.0, seed=0))

    _, _, dao_phot_obj = make_psf_photometry_objs()

    match = 'The fit may be unsuccessful'
    ctx1 = pytest.warns(AstropyUserWarning, match=match)
    if PYTEST_LT_80:
        ctx2 = nullcontext()
    else:
        ctx2 = pytest.warns(AstropyDeprecationWarning)
    with ctx1, ctx2:
        result_table = dao_phot_obj(image)
        result_table['x'] = result_table['x_fit']
        result_table['y'] = result_table['y_fit']
        result_table['flux'] = result_table['flux_fit']

    match1 = 'Both init_guesses and finder are different than None'
    ctx1 = pytest.warns(AstropyUserWarning, match=match1)
    if PYTEST_LT_80:
        ctx2 = nullcontext()
        ctx3 = nullcontext()
        ctx4 = nullcontext()
    else:
        match2 = 'init_guesses contains a "group_id" column'
        match3 = 'The fit may be unsuccessful; check fit_info'
        ctx2 = pytest.warns(AstropyUserWarning, match=match2)
        ctx3 = pytest.warns(AstropyUserWarning, match=match3)
        ctx4 = pytest.warns(AstropyDeprecationWarning)
    with ctx1, ctx2, ctx3, ctx4:
        second_result = dao_phot_obj(image, init_guesses=result_table)
        assert second_result


@pytest.mark.skipif(not HAS_SCIPY, reason='scipy is required')
def test_photometry_mask_nan():
    with pytest.warns(AstropyDeprecationWarning):
        size = 64
        sources1 = Table()
        sources1['flux'] = [800]
        sources1['x_0'] = [size / 2]
        sources1['y_0'] = [size / 2]
        sources1['sigma'] = [6]
        sources1['theta'] = [0]

        img_shape = (size, size)
        data = make_gaussian_prf_sources_image(img_shape, sources1)
        data[30, 20:40] = np.nan

        daogroup = DAOGroup(3.0)
        psf_model = IntegratedGaussianPRF(sigma=2.0)
        psfphot = BasicPSFPhotometry(group_maker=daogroup, finder=None,
                                     bkg_estimator=None, psf_model=psf_model,
                                     fitshape=(11, 11))

        init = Table()
        init['x_0'] = [30]
        init['y_0'] = [30]
        init['flux_0'] = [200.0]

        mask = ~np.isfinite(data)
        tbl = psfphot(data, init_guesses=init, mask=mask)
        assert tbl['x_fit'] != init['x_0']
        assert tbl['y_fit'] != init['y_0']
        assert tbl['flux_fit'] != init['flux_0']

        match = 'Input data contains unmasked non-finite values'
        with pytest.warns(AstropyUserWarning, match=match):
            tbl2 = psfphot(data, init_guesses=init)
            assert tbl == tbl2


@pytest.mark.skipif(not HAS_SCIPY, reason='scipy is required')
def test_photometry_subshape():
    with pytest.warns(AstropyDeprecationWarning):
        size = 21
        cen = (size - 1) // 2
        sigma = 2.0
        sources = Table()
        sources['flux'] = [1]
        sources['x_0'] = [cen]
        sources['y_0'] = [cen]
        sources['sigma'] = [sigma]
        psf = make_gaussian_prf_sources_image((size, size), sources)
        psf_model = FittableImageModel(psf)

        sources = Table()
        sources['flux'] = [2000, 1000]
        sources['x_0'] = [18, 7]
        sources['y_0'] = [17, 25]
        sources['sigma'] = [sigma, sigma]
        shape = (33, 33)
        image = make_gaussian_prf_sources_image(shape, sources)

        daogroup = DAOGroup(crit_separation=8)
        mmm_bkg = MMMBackground()
        iraffind = IRAFStarFinder(threshold=10, fwhm=5, roundlo=-1,
                                  minsep_fwhm=1)
        fitter = LevMarLSQFitter()

        pobj1 = BasicPSFPhotometry(finder=iraffind, group_maker=daogroup,
                                   bkg_estimator=mmm_bkg, psf_model=psf_model,
                                   fitter=fitter, fitshape=(7, 7),
                                   aperture_radius=5, subshape=(3, 3))
        pobj2 = BasicPSFPhotometry(finder=iraffind, group_maker=daogroup,
                                   bkg_estimator=mmm_bkg, psf_model=psf_model,
                                   fitter=fitter, fitshape=(7, 7),
                                   aperture_radius=5, subshape=5)
        pobj3 = BasicPSFPhotometry(finder=iraffind, group_maker=daogroup,
                                   bkg_estimator=mmm_bkg, psf_model=psf_model,
                                   fitter=fitter, fitshape=(7, 7),
                                   aperture_radius=5, subshape=None)
        pobj4 = BasicPSFPhotometry(finder=iraffind, group_maker=daogroup,
                                   bkg_estimator=mmm_bkg, psf_model=psf_model,
                                   fitter=fitter, fitshape=(7, 7),
                                   aperture_radius=5, subshape=7)

        _ = pobj1(image)
        _ = pobj2(image)
        _ = pobj3(image)
        _ = pobj4(image)
        resid1 = pobj1.get_residual_image()
        resid2 = pobj2.get_residual_image()
        resid3 = pobj3.get_residual_image()
        resid4 = pobj4.get_residual_image()
        assert np.sum(resid1) > np.sum(resid2)
        assert_allclose(np.sum(resid3), np.sum(resid4))


@pytest.mark.skipif(not HAS_SCIPY, reason='scipy is required')
def test_subshape_invalid():
    basic_phot_obj = make_psf_photometry_objs()[0]

    with pytest.raises(ValueError):
        basic_phot_obj.subshape = (2, 2)
    with pytest.raises(ValueError):
        basic_phot_obj.subshape = 2
    with pytest.raises(ValueError):
        basic_phot_obj.subshape = (-1, 0)
    with pytest.raises(ValueError):
        basic_phot_obj.subshape = (3, 3, 3)


@pytest.mark.filterwarnings('ignore:Both init_guesses and finder are '
                            'different than None')
@pytest.mark.filterwarnings('ignore:No sources were found')
@pytest.mark.skipif(not HAS_SCIPY, reason='scipy is required')
@pytest.mark.skipif(not minversion(astropy, '5.3'),
                    reason='astropy 5.3 is required')
@pytest.mark.parametrize('sigma_psf, sources',
                         [(sigma_psfs[0], sources1),
                          (sigma_psfs[1], sources2)])
def test_psf_photometry_oneiter_uncert(sigma_psf, sources):
    """
    Make an image with a group of two overlapped stars and an
    isolated one, and check that the best-fit fluxes have smaller
    uncertainties when the measured fluxes have smaller uncertainties.
    """
    img_shape = (32, 32)
    # generate image with read-out noise (Gaussian) and
    # background noise (Poisson)
    image = (make_gaussian_prf_sources_image(img_shape, sources)
             + make_noise_image(img_shape, distribution='poisson', mean=6.0,
                                seed=0)
             + make_noise_image(img_shape, distribution='gaussian', mean=0.0,
                                stddev=2.0, seed=0))

    sigma_clip = SigmaClip(sigma=3.0)
    bkgrms = StdBackgroundRMS(sigma_clip)
    std = bkgrms(image)
    phot_objs = make_psf_photometry_objs(std, sigma_psf)

    flux_uncertainties_0 = []
    flux_uncertainties_1 = []

    for uncertainty_scale_factor, flux_uncert in zip(
        [1e-5, 0.1], [flux_uncertainties_0, flux_uncertainties_1]
    ):
        for phot_proc in phot_objs:
            uncertainty = (
                uncertainty_scale_factor * np.std(image) * np.ones_like(image)
            )
            with pytest.warns(AstropyDeprecationWarning):
                result_tab = phot_proc(image, uncertainty=uncertainty)
            flux_uncert.append(np.array(result_tab['flux_unc']))

    for uncert_0, uncert_1 in zip(flux_uncertainties_0, flux_uncertainties_1):
        assert np.all(uncert_0 < uncert_1)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the model module.
"""

import warnings
from contextlib import nullcontext

import numpy as np
import pytest
from astropy.io import fits
from astropy.utils.data import get_pkg_data_filename

from photutils.datasets import get_path
from photutils.isophote.ellipse import Ellipse
from photutils.isophote.geometry import EllipseGeometry
from photutils.isophote.model import build_ellipse_model
from photutils.isophote.tests.make_test_data import make_test_image
from photutils.tests.helper import PYTEST_LT_80
from photutils.utils._optional_deps import HAS_SCIPY


@pytest.mark.remote_data
@pytest.mark.skipif(not HAS_SCIPY, reason='scipy is required')
def test_model():
    path = get_path('isophote/M105-S001-RGB.fits',
                    location='photutils-datasets', cache=True)
    hdu = fits.open(path)
    data = hdu[0].data[0]
    hdu.close()

    g = EllipseGeometry(530.0, 511, 10.0, 0.1, 10.0 / 180.0 * np.pi)
    ellipse = Ellipse(data, geometry=g, threshold=1.0e5)

    # NOTE: this sometimes emits warnings (e.g., py38, ubuntu), but
    # sometimes not. Here we simply ignore any RuntimeWarning, whether
    # there is one or not.
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', RuntimeWarning)
        isophote_list = ellipse.fit_image()
    model = build_ellipse_model(data.shape, isophote_list,
                                fill=np.mean(data[10:100, 10:100]))

    assert data.shape == model.shape

    residual = data - model
    assert np.mean(residual) <= 5.0
    assert np.mean(residual) >= -5.0


@pytest.mark.skipif(not HAS_SCIPY, reason='scipy is required')
def test_model_simulated_data():
    data = make_test_image(nx=200, ny=200, i0=10.0, sma=5.0, eps=0.5,
                           pa=np.pi / 3.0, noise=0.05, seed=0)

    g = EllipseGeometry(100.0, 100.0, 5.0, 0.5, np.pi / 3.0)
    ellipse = Ellipse(data, geometry=g, threshold=1.0e5)
    isophote_list = ellipse.fit_image()
    model = build_ellipse_model(data.shape, isophote_list,
                                fill=np.mean(data[0:50, 0:50]))

    assert data.shape == model.shape

    residual = data - model
    assert np.mean(residual) <= 5.0
    assert np.mean(residual) >= -5.0


@pytest.mark.skipif(not HAS_SCIPY, reason='scipy is required')
def test_model_minimum_radius():
    # This test requires a "defective" image that drives the
    # model building algorithm into a corner, where it fails.
    # With the algorithm fixed, it bypasses the failure and
    # succeeds in building the model image.
    filepath = get_pkg_data_filename('data/minimum_radius_test.fits')
    with fits.open(filepath) as hdu:
        data = hdu[0].data
        g = EllipseGeometry(50.0, 45, 530.0, 0.1, 10.0 / 180.0 * np.pi)
        g.find_center(data)
        ellipse = Ellipse(data, geometry=g)

        match1 = 'Degrees of freedom'
        ctx1 = pytest.warns(RuntimeWarning, match=match1)
        if PYTEST_LT_80:
            ctx2 = nullcontext()
            ctx3 = nullcontext()
        else:
            match2 = 'Mean of empty slice'
            match3 = 'invalid value encountered'
            ctx2 = pytest.warns(RuntimeWarning, match=match2)
            ctx3 = pytest.warns(RuntimeWarning, match=match3)
        with ctx1, ctx2, ctx3:
            isophote_list = ellipse.fit_image(sma0=40, minsma=0,
                                              maxsma=350.0, step=0.4, nclip=3)

        model = build_ellipse_model(data.shape, isophote_list,
                                    fill=np.mean(data[0:50, 0:50]))

        # It's enough that the algorithm reached this point. The
        # actual accuracy of the modelling is being tested elsewhere.
        assert data.shape == model.shape



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the make module.
"""

import numpy as np
import pytest
from astropy.modeling.models import Moffat2D
from astropy.table import Table
from astropy.utils.exceptions import AstropyUserWarning
from numpy.testing import assert_allclose

from photutils.datasets import (apply_poisson_noise, make_4gaussians_image,
                                make_100gaussians_image,
                                make_gaussian_prf_sources_image,
                                make_gaussian_sources_image, make_gwcs,
                                make_model_sources_image, make_noise_image,
                                make_random_gaussians_table,
                                make_random_models_table, make_test_psf_data,
                                make_wcs)
from photutils.psf import IntegratedGaussianPRF
from photutils.utils._optional_deps import HAS_GWCS, HAS_SCIPY

SOURCE_TABLE = Table()
SOURCE_TABLE['flux'] = [1, 2, 3]
SOURCE_TABLE['x_mean'] = [30, 50, 70.5]
SOURCE_TABLE['y_mean'] = [50, 50, 50.5]
SOURCE_TABLE['x_stddev'] = [1, 2, 3.5]
SOURCE_TABLE['y_stddev'] = [2, 1, 3.5]
SOURCE_TABLE['theta'] = np.array([0.0, 30, 50]) * np.pi / 180.0

SOURCE_TABLE_PRF = Table()
SOURCE_TABLE_PRF['x_0'] = [30, 50, 70.5]
SOURCE_TABLE_PRF['y_0'] = [50, 50, 50.5]
# Without sigma, make_gaussian_prf_sources_image will default to sigma = 1
# so we can ignore it when converting to amplitude
SOURCE_TABLE_PRF['amplitude'] = np.array([1, 2, 3]) / (2 * np.pi)


def test_make_noise_image():
    shape = (100, 100)
    image = make_noise_image(shape, 'gaussian', mean=0.0, stddev=2.0)
    assert image.shape == shape
    assert_allclose(image.mean(), 0.0, atol=1.0)


def test_make_noise_image_poisson():
    shape = (100, 100)
    image = make_noise_image(shape, 'poisson', mean=1.0)
    assert image.shape == shape
    assert_allclose(image.mean(), 1.0, atol=1.0)


def test_make_noise_image_nomean():
    """Test if ValueError raises if mean is not input."""

    with pytest.raises(ValueError):
        shape = (100, 100)
        make_noise_image(shape, 'gaussian', stddev=2.0)


def test_make_noise_image_nostddev():
    """
    Test if ValueError raises if stddev is not input for Gaussian noise.
    """

    with pytest.raises(ValueError):
        shape = (100, 100)
        make_noise_image(shape, 'gaussian', mean=2.0)


def test_apply_poisson_noise():
    shape = (100, 100)
    data = np.ones(shape)
    result = apply_poisson_noise(data)
    assert result.shape == shape
    assert_allclose(result.mean(), 1.0, atol=1.0)


def test_apply_poisson_noise_negative():
    """Test if negative image values raises ValueError."""

    with pytest.raises(ValueError):
        shape = (100, 100)
        data = np.zeros(shape) - 1.0
        apply_poisson_noise(data)


def test_make_gaussian_sources_image():
    shape = (100, 100)
    image = make_gaussian_sources_image(shape, SOURCE_TABLE)
    assert image.shape == shape
    assert_allclose(image.sum(), SOURCE_TABLE['flux'].sum())


@pytest.mark.skipif(not HAS_SCIPY, reason='scipy is required')
def test_make_gaussian_prf_sources_image():
    shape = (100, 100)
    image = make_gaussian_prf_sources_image(shape, SOURCE_TABLE_PRF)
    assert image.shape == shape
    # Without sigma in table, image assumes sigma = 1
    flux = SOURCE_TABLE_PRF['amplitude'] * (2 * np.pi)
    assert_allclose(image.sum(), flux.sum())


def test_make_gaussian_sources_image_amplitude():
    table = SOURCE_TABLE.copy()
    table.remove_column('flux')
    table['amplitude'] = [1, 2, 3]
    shape = (100, 100)
    image = make_gaussian_sources_image(shape, table)
    assert image.shape == shape


def test_make_gaussian_sources_image_oversample():
    shape = (100, 100)
    image = make_gaussian_sources_image(shape, SOURCE_TABLE, oversample=10)
    assert image.shape == shape
    assert_allclose(image.sum(), SOURCE_TABLE['flux'].sum())


def test_make_random_gaussians_table():
    n_sources = 5
    param_ranges = dict([('amplitude', [500, 1000]), ('x_mean', [0, 500]),
                         ('y_mean', [0, 300]), ('x_stddev', [1, 5]),
                         ('y_stddev', [1, 5]), ('theta', [0, np.pi])])

    table = make_random_gaussians_table(n_sources, param_ranges, seed=0)
    assert len(table) == n_sources


def test_make_random_gaussians_table_flux():
    n_sources = 5
    param_ranges = dict([('flux', [500, 1000]), ('x_mean', [0, 500]),
                         ('y_mean', [0, 300]), ('x_stddev', [1, 5]),
                         ('y_stddev', [1, 5]), ('theta', [0, np.pi])])
    table = make_random_gaussians_table(n_sources, param_ranges, seed=0)
    assert 'amplitude' in table.colnames
    assert len(table) == n_sources


def test_make_4gaussians_image():
    shape = (100, 200)
    data_sum = 176219.18059091491
    image = make_4gaussians_image()
    assert image.shape == shape
    assert_allclose(image.sum(), data_sum, rtol=1.0e-6)


def test_make_100gaussians_image():
    shape = (300, 500)
    data_sum = 826182.24501251709
    image = make_100gaussians_image()
    assert image.shape == shape
    assert_allclose(image.sum(), data_sum, rtol=1.0e-6)


def test_make_random_models_table():
    model = Moffat2D(amplitude=1)
    param_ranges = {'x_0': (0, 300), 'y_0': (0, 500),
                    'gamma': (1, 3), 'alpha': (1.5, 3)}
    source_table = make_random_models_table(10, param_ranges)

    # most of the make_model_sources_image options are exercised in the
    # make_gaussian_sources_image tests
    image = make_model_sources_image((300, 500), model, source_table)
    assert image.sum() > 1


def test_make_wcs():
    shape = (100, 200)
    wcs = make_wcs(shape)

    assert wcs.pixel_shape == shape
    assert wcs.wcs.radesys == 'ICRS'

    wcs = make_wcs(shape, galactic=True)
    assert wcs.wcs.ctype[0] == 'GLON-CAR'
    assert wcs.wcs.ctype[1] == 'GLAT-CAR'


@pytest.mark.skipif(not HAS_GWCS, reason='gwcs is required')
def test_make_gwcs():
    shape = (100, 200)

    wcs = make_gwcs(shape)
    assert wcs.pixel_n_dim == 2
    assert wcs.available_frames == ['detector', 'icrs']
    assert wcs.output_frame.name == 'icrs'
    assert wcs.output_frame.axes_names == ('lon', 'lat')

    wcs = make_gwcs(shape, galactic=True)
    assert wcs.pixel_n_dim == 2
    assert wcs.available_frames == ['detector', 'galactic']
    assert wcs.output_frame.name == 'galactic'
    assert wcs.output_frame.axes_names == ('lon', 'lat')


@pytest.mark.skipif(not HAS_GWCS, reason='gwcs is required')
def test_make_wcs_compare():
    shape = (200, 300)
    wcs = make_wcs(shape)
    gwcs_obj = make_gwcs(shape)
    sc1 = wcs.pixel_to_world((50, 75), (50, 100))
    sc2 = gwcs_obj.pixel_to_world((50, 75), (50, 100))

    assert_allclose(sc1.ra, sc2.ra)
    assert_allclose(sc1.dec, sc2.dec)


@pytest.mark.skipif(not HAS_SCIPY, reason='scipy is required')
def test_make_test_psf_data():
    psf_model = IntegratedGaussianPRF(flux=100, sigma=1.5)
    psf_shape = (5, 5)
    nsources = 10
    shape = (100, 100)
    data, true_params = make_test_psf_data(shape, psf_model, psf_shape,
                                           nsources, flux_range=(500, 1000),
                                           min_separation=10, seed=0)

    assert isinstance(data, np.ndarray)
    assert data.shape == shape
    assert isinstance(true_params, Table)
    assert len(true_params) == nsources
    assert true_params['x'].min() >= 0
    assert true_params['y'].min() >= 0

    match = 'Unable to produce'
    with pytest.warns(AstropyUserWarning, match=match):
        nsources = 100
        make_test_psf_data(shape, psf_model, psf_shape, nsources,
                           flux_range=(500, 1000), min_separation=100, seed=0)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the photometry module.
"""

import astropy.units as u
import numpy as np
import pytest
from astropy.modeling.fitting import LMLSQFitter, SimplexLSQFitter
from astropy.modeling.models import Gaussian2D
from astropy.nddata import NDData, StdDevUncertainty
from astropy.table import QTable, Table
from astropy.utils.exceptions import (AstropyDeprecationWarning,
                                      AstropyUserWarning)
from numpy.testing import assert_allclose, assert_equal

from photutils.background import LocalBackground, MMMBackground
from photutils.datasets import (make_gaussian_prf_sources_image,
                                make_noise_image, make_test_psf_data)
from photutils.detection import DAOStarFinder
from photutils.psf import (IntegratedGaussianPRF, IterativePSFPhotometry,
                           PSFPhotometry, SourceGrouper, make_psf_model)
from photutils.psf.photometry_depr import DAOGroup
from photutils.utils._optional_deps import HAS_SCIPY
from photutils.utils.exceptions import NoDetectionsWarning


@pytest.mark.skipif(not HAS_SCIPY, reason='scipy is required')
def test_inputs():
    model = IntegratedGaussianPRF(sigma=1.0)

    match = 'psf_model must be an Astropy Model subclass'
    with pytest.raises(TypeError, match=match):
        _ = PSFPhotometry(1, 3)

    match = 'fit_shape must have an odd value for both axes'
    for shape in ((0, 0), (4, 3)):
        with pytest.raises(ValueError, match=match):
            _ = PSFPhotometry(model, shape)

    match = 'fit_shape must be > 0'
    with pytest.raises(ValueError, match=match):
        _ = PSFPhotometry(model, (-1, 1))

    match = 'fit_shape must be a finite value'
    for shape in ((np.nan, 3), (5, np.inf)):
        with pytest.raises(ValueError, match=match):
            _ = PSFPhotometry(model, shape)

    kwargs = {'grouper': 1, 'finder': 1, 'fitter': 1}
    for key, val in kwargs.items():
        match = f"'{key}' must be a callable object"
        with pytest.raises(TypeError, match=match):
            _ = PSFPhotometry(model, 1, **{key: val})

    match = 'Invalid grouper class. Please use SourceGrouper.'
    with pytest.raises(ValueError, match=match):
        with pytest.warns(AstropyDeprecationWarning):
            _ = PSFPhotometry(model, 1, grouper=DAOGroup(1))

    match = 'localbkg_estimator must be a LocalBackground instance'
    with pytest.raises(ValueError, match=match):
        localbkg = MMMBackground()
        _ = PSFPhotometry(model, 1, localbkg_estimator=localbkg)

    match = 'aperture_radius must be a strictly-positive scalar'
    for radius in (0, -1, np.nan, np.inf):
        with pytest.raises(ValueError, match=match):
            _ = PSFPhotometry(model, 1, aperture_radius=radius)

    match = 'data must be a 2D array'
    psfphot = PSFPhotometry(model, (3, 3))
    with pytest.raises(ValueError, match=match):
        _ = psfphot(np.arange(3))

    match = 'data and mask must have the same shape.'
    with pytest.raises(ValueError, match=match):
        data = np.ones((11, 11))
        mask = np.ones((3, 3))
        _ = psfphot(data, mask=mask)

    match = 'init_params must be an astropy Table'
    with pytest.raises(TypeError, match=match):
        data = np.ones((11, 11))
        _ = psfphot(data, init_params=1)

    match = ('init_param must contain valid column names for the x and y '
             'source positions')
    with pytest.raises(ValueError, match=match):
        tbl = Table()
        tbl['a'] = np.arange(3)
        data = np.ones((11, 11))
        _ = psfphot(data, init_params=tbl)

    # test no finder or init_params
    match = 'finder must be defined if init_params is not input'
    psfphot = PSFPhotometry(model, (3, 3), aperture_radius=5)
    with pytest.raises(ValueError, match=match):
        data = np.ones((11, 11))
        _ = psfphot(data)

    # data has unmasked non-finite value
    match = 'Input data contains unmasked non-finite values'
    psfphot2 = PSFPhotometry(model, (3, 3), aperture_radius=3)
    init_params = Table()
    init_params['x_init'] = [1, 2]
    init_params['y_init'] = [1, 2]
    with pytest.warns(AstropyUserWarning, match=match):
        data = np.ones((11, 11))
        data[5, 5] = np.nan
        _ = psfphot2(data, init_params=init_params)

    # mask is input, but data has unmasked non-finite value
    match = 'Input data contains unmasked non-finite values'
    with pytest.warns(AstropyUserWarning, match=match):
        data = np.ones((11, 11))
        data[5, 5] = np.nan
        mask = np.zeros(data.shape, dtype=bool)
        mask[7, 7] = True
        _ = psfphot2(data, mask=mask, init_params=init_params)

    # this should not raise a warning because the non-finite pixel was
    # explicitly masked
    data = np.ones((11, 11))
    data[5, 5] = np.nan
    mask = np.zeros(data.shape, dtype=bool)
    mask[5, 5] = True
    _ = psfphot2(data, mask=mask, init_params=init_params)


@pytest.fixture(name='test_data')
def fixture_test_data():
    psf_model = IntegratedGaussianPRF(flux=1, sigma=2.7 / 2.35)
    psf_shape = (9, 9)
    nsources = 10
    shape = (101, 101)
    data, true_params = make_test_psf_data(shape, psf_model, psf_shape,
                                           nsources, flux_range=(500, 700),
                                           min_separation=10, seed=0)
    noise = make_noise_image(data.shape, mean=0, stddev=1, seed=0)
    data += noise
    error = np.abs(noise)

    return data, error, true_params


@pytest.mark.skipif(not HAS_SCIPY, reason='scipy is required')
def test_psf_photometry(test_data):
    data, error, sources = test_data

    psf_model = IntegratedGaussianPRF(flux=1, sigma=2.7 / 2.35)
    fit_shape = (5, 5)
    finder = DAOStarFinder(6.0, 2.0)
    psfphot = PSFPhotometry(psf_model, fit_shape, finder=finder,
                            aperture_radius=4)
    phot = psfphot(data, error=error)
    resid_data = psfphot.make_residual_image(data, fit_shape)

    assert isinstance(phot, QTable)
    assert len(phot) == len(sources)
    assert isinstance(resid_data, np.ndarray)
    assert resid_data.shape == data.shape

    unit = u.Jy
    photu = psfphot(data * unit, error=error * unit)
    assert photu['flux_init'].unit == unit
    assert photu['flux_fit'].unit == unit
    assert photu['flux_err'].unit == unit
    resid_datau = psfphot.make_residual_image(data * u.Jy, fit_shape)
    assert resid_datau.unit == unit

    photm = psfphot(data, error=error, mask=np.ma.nomask)
    assert np.all(phot == photm)

    # test NDData input
    uncertainty = StdDevUncertainty(error)
    nddata = NDData(data, uncertainty=uncertainty)
    psfphot2 = PSFPhotometry(psf_model, fit_shape, finder=finder,
                             aperture_radius=4)
    phot2 = psfphot2(nddata)
    resid_data2 = psfphot2.make_residual_image(nddata, fit_shape)

    assert np.all(phot == phot2)
    assert isinstance(resid_data2, NDData)
    assert resid_data2.data.shape == data.shape
    assert_allclose(resid_data, resid_data2.data)

    # test NDData input with units
    unit = u.Jy
    uncertainty = StdDevUncertainty(error)
    nddata = NDData(data, uncertainty=uncertainty, unit=unit)
    psfphotu = PSFPhotometry(psf_model, fit_shape, finder=finder,
                             aperture_radius=4)
    photu = psfphotu(nddata)
    assert photu['flux_init'].unit == unit
    assert photu['flux_fit'].unit == unit
    assert photu['flux_err'].unit == unit
    resid_data3 = psfphotu.make_residual_image(nddata, fit_shape)
    assert resid_data3.unit == unit


@pytest.mark.skipif(not HAS_SCIPY, reason='scipy is required')
def test_psf_photometry_mask(test_data):
    data, error, sources = test_data
    data_orig = data.copy()
    data = data.copy()
    data[50, 40:50] = np.nan

    psf_model = IntegratedGaussianPRF(flux=1, sigma=2.7 / 2.35)
    fit_shape = (5, 5)
    finder = DAOStarFinder(6.0, 2.0)
    psfphot = PSFPhotometry(psf_model, fit_shape, finder=finder,
                            aperture_radius=4)

    match = 'Input data contains unmasked non-finite values'
    with pytest.warns(AstropyUserWarning, match=match):
        phot = psfphot(data, error=error, mask=None)
        assert len(phot) == len(sources)

    # unmasked NaN with mask not None
    match = 'Input data contains unmasked non-finite values'
    with pytest.warns(AstropyUserWarning, match=match):
        mask = ~np.isfinite(data)
        mask[50, 40] = False
        phot = psfphot(data, error=error, mask=mask)
        assert len(phot) == len(sources)

    mask = ~np.isfinite(data)
    phot2 = psfphot(data, error=error, mask=mask)
    assert np.all(phot == phot2)

    # mask all True; finder returns no sources
    with pytest.warns(NoDetectionsWarning):
        mask = np.ones(data.shape, dtype=bool)
        _ = psfphot(data, mask=mask)

    # completely masked source
    match = ('is completely masked. Remove the source from init_params '
             'or correct the input mask')
    with pytest.raises(ValueError, match=match):
        init_params = QTable()
        init_params['x'] = [42]
        init_params['y'] = [36]
        mask = np.ones(data.shape, dtype=bool)
        _ = psfphot(data, mask=mask, init_params=init_params)

    # completely masked source
    match = ('The number of data points is less than the number of fit '
             'parameters.')
    with pytest.raises(ValueError, match=match):
        init_params = QTable()
        init_params['x'] = [42]
        init_params['y'] = [36]
        mask = np.zeros(data.shape, dtype=bool)
        mask[35:37, :] = True
        mask[37, 42:44] = True
        psfphot = PSFPhotometry(psf_model, (3, 3), finder=finder,
                                aperture_radius=4)
        _ = psfphot(data_orig, mask=mask, init_params=init_params)

    # masked central pixel
    init_params = QTable()
    init_params['x'] = [42]
    init_params['y'] = [36]
    mask = np.zeros(data.shape, dtype=bool)
    mask[36, 42] = True
    phot = psfphot(data_orig, mask=mask, init_params=init_params)
    assert len(phot) == 1


@pytest.mark.skipif(not HAS_SCIPY, reason='scipy is required')
def test_psf_photometry_init_params(test_data):
    data, error, _ = test_data
    data = data.copy()

    psf_model = IntegratedGaussianPRF(flux=1, sigma=2.7 / 2.35)
    fit_shape = (5, 5)
    finder = DAOStarFinder(6.0, 2.0)
    psfphot = PSFPhotometry(psf_model, fit_shape, finder=finder,
                            aperture_radius=4)

    init_params = QTable()
    init_params['x'] = [42]
    init_params['y'] = [36]
    phot = psfphot(data, error=error, init_params=init_params)
    assert isinstance(phot, QTable)
    assert len(phot) == 1

    match = 'aperture_radius must be defined if init_params is not input'
    with pytest.raises(ValueError, match=match):
        psfphot = PSFPhotometry(psf_model, fit_shape, finder=finder,
                                aperture_radius=None)
        _ = psfphot(data, error=error, init_params=init_params)

    init_params['flux'] = 650
    phot = psfphot(data, error=error, init_params=init_params)
    assert len(phot) == 1

    init_params['group_id'] = 1
    phot = psfphot(data, error=error, init_params=init_params)
    assert len(phot) == 1

    init_params['flux'] = [650 * u.Jy]
    match = ('init_params flux column has units, but the input data does '
             'not have units')
    with pytest.raises(ValueError, match=match):
        _ = psfphot(data, error=error, init_params=init_params)

    init_params['flux'] = [650 * u.Jy]
    match = ('init_params flux column has units that are incompatible with '
             'the input data units')
    with pytest.raises(ValueError, match=match):
        _ = psfphot(data << u.m, init_params=init_params)

    init_params['flux'] = [650]
    match = ('The input data has units, but the init_params flux column '
             'does not have units')
    with pytest.raises(ValueError, match=match):
        _ = psfphot(data << u.Jy, init_params=init_params)

    init_params = QTable()
    init_params['x'] = [-42]
    init_params['y'] = [-36]
    init_params['flux'] = [100]
    match = 'does not overlap with the input data'
    with pytest.raises(ValueError, match=match):
        _ = psfphot(data, init_params=init_params)


@pytest.mark.skipif(not HAS_SCIPY, reason='scipy is required')
def test_psf_photometry_init_params_columns(test_data):
    data, error, _ = test_data
    data = data.copy()

    psf_model = IntegratedGaussianPRF(flux=1, sigma=2.7 / 2.35)
    fit_shape = (5, 5)
    finder = DAOStarFinder(6.0, 2.0)
    psfphot = PSFPhotometry(psf_model, fit_shape, finder=finder)

    xy_suffixes = ('_init', 'init', 'centroid', '_centroid', '_peak', '',
                   'cen', '_cen', 'pos', '_pos', '_0', '0')
    flux_cols = ['flux_init', 'flux_0', 'flux0', 'flux', 'source_sum',
                 'segment_flux', 'kron_flux']
    flux_cols += flux_cols[0:4]  # pad to have same length as xy_suffixes

    xcols = ['x' + i for i in xy_suffixes]
    ycols = ['y' + i for i in xy_suffixes]

    phots = []
    for xcol, ycol, fluxcol in zip(xcols, ycols, flux_cols):
        init_params = QTable()
        init_params[xcol] = [42]
        init_params[ycol] = [36]
        init_params[fluxcol] = [680]
        phot = psfphot(data, error=error, init_params=init_params)
        assert isinstance(phot, QTable)
        assert len(phot) == 1
        phots.append(phot)

    for phot in phots[1:]:
        assert_allclose(phot['x_fit'], phots[0]['x_fit'])
        assert_allclose(phot['y_fit'], phots[0]['y_fit'])
        assert_allclose(phot['flux_fit'], phots[0]['flux_fit'])


@pytest.mark.skipif(not HAS_SCIPY, reason='scipy is required')
def test_grouper(test_data):
    data, error, sources = test_data

    psf_model = IntegratedGaussianPRF(flux=1, sigma=2.7 / 2.35)
    fit_shape = (5, 5)
    finder = DAOStarFinder(6.0, 2.0)
    grouper = SourceGrouper(min_separation=20)
    psfphot = PSFPhotometry(psf_model, fit_shape, finder=finder,
                            grouper=grouper, aperture_radius=4)
    phot = psfphot(data, error=error)
    assert isinstance(phot, QTable)
    assert len(phot) == len(sources)
    assert_equal(phot['group_id'], (1, 2, 3, 4, 5, 5, 5, 6, 6, 7))
    assert_equal(phot['group_size'], (1, 1, 1, 1, 3, 3, 3, 2, 2, 1))


@pytest.mark.skipif(not HAS_SCIPY, reason='scipy is required')
def test_large_group_warning():
    psf_model = IntegratedGaussianPRF(flux=1, sigma=1.0)
    grouper = SourceGrouper(min_separation=50)
    psf_shape = (5, 5)
    fit_shape = (5, 5)
    nsources = 50
    shape = (301, 301)
    data, true_params = make_test_psf_data(shape, psf_model, psf_shape,
                                           nsources, flux_range=(500, 700),
                                           min_separation=10, seed=0)
    match = 'Some groups have more than'
    with pytest.warns(AstropyUserWarning, match=match):
        psfphot = PSFPhotometry(psf_model, fit_shape, grouper=grouper)
        psfphot(data, init_params=true_params)


@pytest.mark.skipif(not HAS_SCIPY, reason='scipy is required')
def test_local_bkg(test_data):
    data, error, sources = test_data

    psf_model = IntegratedGaussianPRF(flux=1, sigma=2.7 / 2.35)
    fit_shape = (5, 5)
    finder = DAOStarFinder(6.0, 2.0)
    grouper = SourceGrouper(min_separation=20)
    bkgstat = MMMBackground()
    localbkg_estimator = LocalBackground(5, 10, bkgstat)
    finder = DAOStarFinder(10.0, 2.0)

    psfphot = PSFPhotometry(psf_model, fit_shape, finder=finder,
                            grouper=grouper, aperture_radius=4,
                            localbkg_estimator=localbkg_estimator)
    phot = psfphot(data, error=error)
    assert np.count_nonzero(phot['local_bkg']) == len(sources)


@pytest.mark.skipif(not HAS_SCIPY, reason='scipy is required')
def test_fixed_params(test_data):
    data, error, _ = test_data

    psf_model = IntegratedGaussianPRF(flux=1, sigma=2.7 / 2.35)
    psf_model.x_0.fixed = True
    psf_model.y_0.fixed = True
    psf_model.flux.fixed = True
    fit_shape = (5, 5)
    finder = DAOStarFinder(6.0, 2.0)
    psfphot = PSFPhotometry(psf_model, fit_shape, finder=finder,
                            aperture_radius=4)

    match = r'One or more fit\(s\) may not have converged.'
    with pytest.warns(AstropyUserWarning, match=match):
        phot = psfphot(data, error=error)
        assert np.all(np.isnan(phot['x_err']))
        assert np.all(np.isnan(phot['y_err']))
        assert np.all(np.isnan(phot['flux_err']))


@pytest.mark.skipif(not HAS_SCIPY, reason='scipy is required')
def test_fit_warning(test_data):
    data, _, _ = test_data

    psf_model = IntegratedGaussianPRF(flux=1, sigma=2.7 / 2.35)
    psf_model.flux.fixed = False
    fit_shape = (5, 5)
    fitter = LMLSQFitter()  # uses "status" instead of "ierr"
    finder = DAOStarFinder(6.0, 2.0)
    # set fitter_maxiters = 1 so that the fit error status is set
    psfphot = PSFPhotometry(psf_model, fit_shape, fitter=fitter,
                            fitter_maxiters=1, finder=finder,
                            aperture_radius=4)

    match = r'One or more fit\(s\) may not have converged.'
    with pytest.warns(AstropyUserWarning, match=match):
        _ = psfphot(data)
        assert len(psfphot.fit_results['fit_error_indices']) > 0


@pytest.mark.skipif(not HAS_SCIPY, reason='scipy is required')
def test_fitter_no_maxiters_no_residuals(test_data):
    data, error, _ = test_data

    psf_model = IntegratedGaussianPRF(flux=1, sigma=2.7 / 2.35)
    psf_model.flux.fixed = False
    fit_shape = (5, 5)
    fitter = SimplexLSQFitter()  # does not produce residual array
    finder = DAOStarFinder(6.0, 2.0)
    match = '"maxiters" will be ignored because it is not accepted by'
    with pytest.warns(AstropyUserWarning, match=match):
        psfphot = PSFPhotometry(psf_model, fit_shape, fitter=fitter,
                                finder=finder, aperture_radius=4)
        phot = psfphot(data, error=error)
        assert np.all(np.isnan(phot['qfit']))
        assert np.all(np.isnan(phot['cfit']))


@pytest.mark.skipif(not HAS_SCIPY, reason='scipy is required')
def test_iterative_psf_photometry(test_data):
    data, error, sources = test_data

    psf_model = IntegratedGaussianPRF(flux=1, sigma=2.7 / 2.35)
    fit_shape = (5, 5)
    bkgstat = MMMBackground()
    localbkg_estimator = LocalBackground(5, 10, bkgstat)
    finder = DAOStarFinder(10.0, 2.0)

    init_params = QTable()
    init_params['x'] = [54, 29, 80]
    init_params['y'] = [8, 26, 29]
    psfphot = IterativePSFPhotometry(psf_model, fit_shape, finder=finder,
                                     localbkg_estimator=localbkg_estimator,
                                     aperture_radius=4)
    phot = psfphot(data, error=error, init_params=init_params)

    assert 'iter_detected' in phot.colnames
    assert len(phot) == len(sources)

    resid_data = psfphot.make_residual_image(data, fit_shape)
    assert isinstance(resid_data, np.ndarray)
    assert resid_data.shape == data.shape

    nddata = NDData(data)
    resid_nddata = psfphot.make_residual_image(nddata, fit_shape)
    assert isinstance(resid_nddata, NDData)
    assert resid_nddata.data.shape == data.shape

    unit = u.Jy
    resid_data = psfphot.make_residual_image(data * unit, fit_shape)
    assert resid_data.unit == unit

    nddata = NDData(data * unit)
    resid_nddata = psfphot.make_residual_image(nddata, fit_shape)
    assert isinstance(resid_nddata, NDData)
    assert resid_nddata.unit == unit

    # test return None if no stars are found on first iteration
    finder = DAOStarFinder(1000.0, 2.0)
    psfphot = IterativePSFPhotometry(psf_model, fit_shape, finder=finder,
                                     localbkg_estimator=localbkg_estimator,
                                     aperture_radius=4)
    with pytest.warns(NoDetectionsWarning):
        phot = psfphot(data, error=error)
        assert phot is None


@pytest.mark.skipif(not HAS_SCIPY, reason='scipy is required')
def test_iterative_psf_photometry_inputs():
    psf_model = IntegratedGaussianPRF(flux=1, sigma=2.7 / 2.35)
    fit_shape = (5, 5)
    finder = DAOStarFinder(10.0, 2.0)

    match = 'finder cannot be None for IterativePSFPhotometry'
    with pytest.raises(ValueError, match=match):
        _ = IterativePSFPhotometry(psf_model, fit_shape, finder=None,
                                   aperture_radius=4)

    match = 'aperture_radius cannot be None for IterativePSFPhotometry'
    with pytest.raises(ValueError, match=match):
        _ = IterativePSFPhotometry(psf_model, fit_shape, finder=finder,
                                   aperture_radius=None)

    match = 'maxiters must be a strictly-positive scalar'
    with pytest.raises(ValueError, match=match):
        _ = IterativePSFPhotometry(psf_model, fit_shape, finder=finder,
                                   aperture_radius=4, maxiters=-1)
    with pytest.raises(ValueError, match=match):
        _ = IterativePSFPhotometry(psf_model, fit_shape, finder=finder,
                                   aperture_radius=4, maxiters=[1, 2])

    match = 'maxiters must be an integer'
    with pytest.raises(ValueError, match=match):
        _ = IterativePSFPhotometry(psf_model, fit_shape, finder=finder,
                                   aperture_radius=4, maxiters=3.14)


@pytest.mark.skipif(not HAS_SCIPY, reason='scipy is required')
def test_negative_xy():
    sources = Table()
    sources['id'] = np.arange(3) + 1
    sources['flux'] = 1
    sources['x_0'] = [-1.4, 15.2, -0.7]
    sources['y_0'] = [-0.3, -0.4, 18.7]
    sources['sigma'] = 3.1
    shape = (31, 31)
    data = make_gaussian_prf_sources_image(shape, sources)
    psf_model = IntegratedGaussianPRF(flux=1, sigma=3.1)
    fit_shape = (11, 11)
    psfphot = PSFPhotometry(psf_model, fit_shape,
                            aperture_radius=10)
    phot = psfphot(data, init_params=sources)
    assert_equal(phot['x_init'], sources['x_0'])
    assert_equal(phot['y_init'], sources['y_0'])


@pytest.mark.skipif(not HAS_SCIPY, reason='scipy is required')
def test_out_of_bounds_centroids():
    sources = Table()
    sources['id'] = np.arange(8) + 1
    sources['flux'] = 1
    sources['x_0'] = [-1.4, 34.5, 14.2, -0.7, 34.5, 14.2, 51.3, 52.0]
    sources['y_0'] = [13, -0.2, -1.6, 40, 51.1, 50.9, 12.2, 42.3]
    sources['sigma'] = 3.1

    shape = (51, 51)
    data = make_gaussian_prf_sources_image(shape, sources)

    psf_model = IntegratedGaussianPRF(flux=1, sigma=3.1)
    fit_shape = (11, 11)
    psfphot = PSFPhotometry(psf_model, fit_shape,
                            aperture_radius=10)

    phot = psfphot(data, init_params=sources)

    # at least one of the best-fit centroids should be
    # out of the bounds of the dataset, producing a
    # masked value in the `cfit` column:
    assert np.any(np.isnan(phot['cfit']))


@pytest.mark.skipif(not HAS_SCIPY, reason='scipy is required')
def test_make_psf_model():
    normalize = False
    sigma = 3.0
    amplitude = 1.0 / (2 * np.pi * sigma**2)
    xcen = ycen = 0.0
    psf0 = Gaussian2D(amplitude, xcen, ycen, sigma, sigma)
    psf1 = make_psf_model(psf0, x_name='x_mean', y_name='y_mean',
                          normalize=normalize)
    psf2 = make_psf_model(psf0, normalize=normalize)
    psf3 = make_psf_model(psf0, x_name='x_mean', normalize=normalize)
    psf4 = make_psf_model(psf0, y_name='y_mean', normalize=normalize)

    yy, xx = np.mgrid[0:101, 0:101]
    psf = psf1.copy()
    xval = 48
    yval = 52
    flux = 14.51
    psf.x_mean_2 = xval
    psf.y_mean_2 = yval
    data = psf(xx, yy) * flux

    fit_shape = 7
    init_params = Table([[46.1], [57.3], [7.1]],
                        names=['x_0', 'y_0', 'flux_0'])
    phot1 = PSFPhotometry(psf1, fit_shape, aperture_radius=None)
    tbl1 = phot1(data, init_params=init_params)

    phot2 = PSFPhotometry(psf2, fit_shape, aperture_radius=None)
    tbl2 = phot2(data, init_params=init_params)

    phot3 = PSFPhotometry(psf3, fit_shape, aperture_radius=None)
    tbl3 = phot3(data, init_params=init_params)

    phot4 = PSFPhotometry(psf4, fit_shape, aperture_radius=None)
    tbl4 = phot4(data, init_params=init_params)

    assert_allclose((tbl1['x_fit'][0], tbl1['y_fit'][0],
                     tbl1['flux_fit'][0]), (xval, yval, flux))
    assert_allclose((tbl2['x_fit'][0], tbl2['y_fit'][0],
                     tbl2['flux_fit'][0]), (xval, yval, flux))
    assert_allclose((tbl3['x_fit'][0], tbl3['y_fit'][0],
                     tbl3['flux_fit'][0]), (xval, yval, flux))
    assert_allclose((tbl4['x_fit'][0], tbl4['y_fit'][0],
                     tbl4['flux_fit'][0]), (xval, yval, flux))



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This module provides tools for generating footprints.
"""

import numpy as np

__all__ = ['circular_footprint']


def circular_footprint(radius, dtype=int):
    """
    Create a circular footprint.

    A pixel is considered to be entirely in or out of the footprint
    depending on whether its center is in or out of the footprint.  The
    size of the output array is the minimal bounding box for the
    footprint.

    Parameters
    ----------
    radius : int
        The radius of the circular footprint.

    dtype : data-type, optional
        The data type of the output `~numpy.ndarray`.

    Returns
    -------
    footprint : `~numpy.ndarray`
        A footprint where array elements are 1 within the footprint and
        0 otherwise.

    Examples
    --------
    >>> from photutils.utils import circular_footprint
    >>> circular_footprint(2)
    array([[0, 0, 1, 0, 0],
           [0, 1, 1, 1, 0],
           [1, 1, 1, 1, 1],
           [0, 1, 1, 1, 0],
           [0, 0, 1, 0, 0]])
    """
    if ~np.isfinite(radius) or radius <= 0 or int(radius) != radius:
        raise ValueError('radius must be a positive, finite integer greater '
                         'than 0')

    x = np.arange(-radius, radius + 1)
    xx, yy = np.meshgrid(x, x)
    return np.array((xx**2 + yy**2) <= radius**2, dtype=dtype)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This module defines classes to estimate local background using a
circular annulus aperture.
"""

import numpy as np

from photutils.aperture import CircularAnnulus
from photutils.background import MedianBackground

__all__ = ['LocalBackground']


class LocalBackground:
    """
    Class to compute a local background using a circular annulus
    aperture.

    Parameters
    ----------
    inner_radius : float
        The inner radius of the circular annulus in pixels.

    outer_radius : float
        The outer radius of the circular annulus in pixels.

    bkg_estimator : callable, optional
        A callable object (a function or e.g., an instance of any
        `~photutils.background.BackgroundBase` subclass) used to
        estimate the background in each aperture. The callable object
        must take in a 2D `~numpy.ndarray` or `~numpy.ma.MaskedArray`
        and have an ``axis`` keyword. The default is an instance of
        `~photutils.background.MedianBackground` with sigma clipping
        (i.e., sigma-clipped median).
    """
    def __init__(self, inner_radius, outer_radius,
                 bkg_estimator=MedianBackground()):
        self.inner_radius = inner_radius
        self.outer_radius = outer_radius
        self.bkg_estimator = bkg_estimator
        self._aperture = CircularAnnulus((0, 0), inner_radius, outer_radius)

    def __call__(self, data, x, y, mask=None):
        """
        Measure the local background in a circular annulus.

        Parameters
        ----------
        data : 2D `~numpy.ndarray`
            The 2D array on which to measure the local background.

        x, y : float or 1D float `~numpy.ndarray`
            The aperture center (x, y) position(s) at which to measure
            the local background.

        mask : 2D bool `~numpy.ndarray`, optional
            A boolean mask with the same shape as ``data`` where a
            `True` value indicates the corresponding element of ``data``
            is masked. Masked data are excluded from all calculations.

        Returns
        -------
        value : float or 1D float `~numpy.ndarray`
            The local background values.
        """
        x = np.atleast_1d(x)
        y = np.atleast_1d(y)

        self._aperture.positions = np.array(list(zip(x, y)))
        apermasks = self._aperture.to_mask(method='center')

        bkg = []
        for apermask in apermasks:
            values = apermask.get_values(data, mask=mask)
            bkg.append(self.bkg_estimator(values))
        bkg = np.array(bkg)

        if bkg.size == 1:
            bkg = bkg[0]

        return bkg



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
The module contains tools for centroiding sources.
"""

import inspect
import warnings

import numpy as np
from astropy.nddata.utils import overlap_slices
from astropy.utils.exceptions import AstropyUserWarning

from photutils.utils._parameters import as_pair
from photutils.utils._round import py2intround

__all__ = ['centroid_com', 'centroid_quadratic', 'centroid_sources']


def centroid_com(data, mask=None):
    """
    Calculate the centroid of an n-dimensional array as its "center of
    mass" determined from moments.

    Non-finite values (e.g., NaN or inf) in the ``data`` array are
    automatically masked.

    Parameters
    ----------
    data : `~numpy.ndarray`
        The input n-dimensional array. The image should be a
        background-subtracted cutout image containing a single source.

    mask : bool `~numpy.ndarray`, optional
        A boolean mask, with the same shape as ``data``, where a `True`
        value indicates the corresponding element of ``data`` is masked.

    Returns
    -------
    centroid : `~numpy.ndarray`
        The coordinates of the centroid in pixel order (e.g., ``(x, y)``
        or ``(x, y, z)``), not numpy axis order.
    """
    # preserve input data - which should be a small cutout image
    data = data.copy()

    if mask is not None and mask is not np.ma.nomask:
        mask = np.asarray(mask, dtype=bool)
        if data.shape != mask.shape:
            raise ValueError('data and mask must have the same shape.')
        data[mask] = 0.0

    badmask = ~np.isfinite(data)
    if np.any(badmask):
        warnings.warn('Input data contains non-finite values (e.g., NaN or '
                      'inf) that were automatically masked.',
                      AstropyUserWarning)
        data[badmask] = 0.0

    total = np.sum(data)
    if total == 0:
        return np.array((np.nan, np.nan))

    indices = np.ogrid[tuple(slice(0, i) for i in data.shape)]

    # note the output array is reversed to give (x, y) order
    return np.array([np.sum(indices[axis] * data) / total
                     for axis in range(data.ndim)])[::-1]


def centroid_quadratic(data, xpeak=None, ypeak=None, fit_boxsize=5,
                       search_boxsize=None, mask=None):
    """
    Calculate the centroid of an n-dimensional array by fitting a 2D
    quadratic polynomial.

    A second degree 2D polynomial is fit within a small region of the
    data defined by ``fit_boxsize`` to calculate the centroid position.
    The initial center of the fitting box can specified using the
    ``xpeak`` and ``ypeak`` keywords. If both ``xpeak`` and ``ypeak``
    are `None`, then the box will be centered at the position of the
    maximum value in the input ``data``.

    If ``xpeak`` and ``ypeak`` are specified, the ``search_boxsize``
    optional keyword can be used to further refine the initial center of
    the fitting box by searching for the position of the maximum pixel
    within a box of size ``search_boxsize``.

    `Vakili & Hogg (2016) <https://arxiv.org/abs/1610.05873>`_
    demonstrate that 2D quadratic centroiding comes very
    close to saturating the `Cramér-Rao lower bound
    <https://en.wikipedia.org/wiki/Cram%C3%A9r%E2%80%93Rao_bound>`_ in a
    wide range of conditions.

    Parameters
    ----------
    data : 2D `~numpy.ndarray`
        The 2D image data. The image should be a background-subtracted
        cutout image containing a single source.

    xpeak, ypeak : float or `None`, optional
        The initial guess of the position of the centroid. If either
        ``xpeak`` or ``ypeak`` is `None` then the position of the
        maximum value in the input ``data`` will be used as the initial
        guess.

    fit_boxsize : int or tuple of int, optional
        The size (in pixels) of the box used to define the fitting
        region. If ``fit_boxsize`` has two elements, they must be in
        ``(ny, nx)`` order. If ``fit_boxsize`` is a scalar then a square
        box of size ``fit_boxsize`` will be used. ``fit_boxsize`` must
        have odd values for both axes.

    search_boxsize : int or tuple of int, optional
        The size (in pixels) of the box used to search for the maximum
        pixel value if ``xpeak`` and ``ypeak`` are both specified. If
        ``fit_boxsize`` has two elements, they must be in ``(ny,
        nx)`` order. If ``search_boxsize`` is a scalar then a square
        box of size ``search_boxsize`` will be used. ``search_boxsize``
        must have odd values for both axes. This parameter is ignored
        if either ``xpeak`` or ``ypeak`` is `None`. In that case, the
        entire array is searched for the maximum value.

    mask : bool `~numpy.ndarray`, optional
        A boolean mask, with the same shape as ``data``, where a `True`
        value indicates the corresponding element of ``data`` is masked.
        Masked data are excluded from calculations.

    Returns
    -------
    centroid : `~numpy.ndarray`
        The ``x, y`` coordinates of the centroid.

    Notes
    -----
    Use ``fit_boxsize = (3, 3)`` to match the work of `Vakili &
    Hogg (2016) <https://arxiv.org/abs/1610.05873>`_ for their 2D
    second-order polynomial centroiding method.

    Because this centroid is based on fitting data, it can fail for many
    reasons, returning (np.nan, np.nan):

        * quadratic fit failed
        * quadratic fit does not have a maximum
        * quadratic fit maximum falls outside image
        * not enough unmasked data points (6 are required)

    Also note that a fit is not performed if the maximum data value is
    at the edge of the data. In this case, the position of the maximum
    pixel will be returned.

    References
    ----------
    .. [1] Vakili and Hogg 2016; arXiv:1610.05873
        (https://arxiv.org/abs/1610.05873)
    """
    if ((xpeak is None and ypeak is not None)
            or (xpeak is not None and ypeak is None)):
        raise ValueError('xpeak and ypeak must both be input or "None"')

    if xpeak is not None and ((xpeak < 0) or (xpeak > data.shape[1] - 1)):
        raise ValueError('xpeak is outside of the input data')
    if ypeak is not None and ((ypeak < 0) or (ypeak > data.shape[0] - 1)):
        raise ValueError('ypeak is outside of the input data')

    # preserve input data - which should be a small cutout image
    data = np.asanyarray(data, dtype=float).copy()
    ny, nx = data.shape

    badmask = ~np.isfinite(data)
    if mask is not None:
        if data.shape != mask.shape:
            raise ValueError('data and mask must have the same shape.')
        data[mask] = np.nan
        badmask &= ~mask

    if np.any(badmask):
        warnings.warn('Input data contains non-finite values (e.g., NaN or '
                      'inf) that were automatically masked.',
                      AstropyUserWarning)
        data[badmask] = np.nan

    fit_boxsize = as_pair('fit_boxsize', fit_boxsize, lower_bound=(0, 1),
                          upper_bound=data.shape, check_odd=True)

    if np.prod(fit_boxsize) < 6:
        raise ValueError('fit_boxsize is too small.  6 values are required '
                         'to fit a 2D quadratic polynomial.')

    if xpeak is None or ypeak is None:
        yidx, xidx = np.unravel_index(np.nanargmax(data), data.shape)
    else:
        xidx = py2intround(xpeak)
        yidx = py2intround(ypeak)

        if search_boxsize is not None:
            search_boxsize = as_pair('search_boxsize', search_boxsize,
                                     lower_bound=(0, 1),
                                     upper_bound=data.shape, check_odd=True)

            slc_data, _ = overlap_slices(data.shape, search_boxsize,
                                         (yidx, xidx), mode='trim')
            cutout = data[slc_data]
            yidx, xidx = np.unravel_index(np.nanargmax(cutout), cutout.shape)
            xidx += slc_data[1].start
            yidx += slc_data[0].start

    # if peak is at the edge of the data, return the position of the maximum
    if xidx == 0 or xidx == nx - 1 or yidx == 0 or yidx == ny - 1:
        warnings.warn('maximum value is at the edge of the data and its '
                      'position was returned; no quadratic fit was '
                      'performed', AstropyUserWarning)
        return np.array((xidx, yidx), dtype=float)

    # extract the fitting region
    slc_data, _ = overlap_slices(data.shape, fit_boxsize, (yidx, xidx),
                                 mode='trim')
    xidx0, xidx1 = (slc_data[1].start, slc_data[1].stop)
    yidx0, yidx1 = (slc_data[0].start, slc_data[0].stop)

    # shift the fitting box if it was clipped by the data edge
    if (xidx1 - xidx0) < fit_boxsize[1]:
        if xidx0 == 0:
            xidx1 = min(nx, xidx0 + fit_boxsize[1])
        if xidx1 == nx:
            xidx0 = max(0, xidx1 - fit_boxsize[1])
    if (yidx1 - yidx0) < fit_boxsize[0]:
        if yidx0 == 0:
            yidx1 = min(ny, yidx0 + fit_boxsize[0])
        if yidx1 == ny:
            yidx0 = max(0, yidx1 - fit_boxsize[0])

    cutout = data[yidx0:yidx1, xidx0:xidx1].ravel()
    if np.count_nonzero(~np.isnan(cutout)) < 6:
        warnings.warn('at least 6 unmasked data points are required to '
                      'perform a 2D quadratic fit',
                      AstropyUserWarning)
        return np.array((np.nan, np.nan))

    # fit a 2D quadratic polynomial to the fitting region
    xi = np.arange(xidx0, xidx1)
    yi = np.arange(yidx0, yidx1)
    x, y = np.meshgrid(xi, yi)
    x = x.ravel()
    y = y.ravel()
    coeff_matrix = np.vstack((np.ones_like(x), x, y, x * y, x * x, y * y)).T

    # remove NaNs from data to be fit
    mask = ~np.isnan(cutout)
    if np.any(mask):
        coeff_matrix = coeff_matrix[mask]
        cutout = cutout[mask]

    try:
        c = np.linalg.lstsq(coeff_matrix, cutout, rcond=None)[0]
    except np.linalg.LinAlgError:
        warnings.warn('quadratic fit failed', AstropyUserWarning)
        return np.array((np.nan, np.nan))

    # analytically find the maximum of the polynomial
    _, c10, c01, c11, c20, c02 = c
    det = 4 * c20 * c02 - c11**2
    if det <= 0 or ((c20 > 0.0 and c02 >= 0.0) or (c20 >= 0.0 and c02 > 0.0)):
        warnings.warn('quadratic fit does not have a maximum',
                      AstropyUserWarning)
        return np.array((np.nan, np.nan))

    xm = (c01 * c11 - 2.0 * c02 * c10) / det
    ym = (c10 * c11 - 2.0 * c20 * c01) / det
    if 0.0 < xm < (nx - 1.0) and 0.0 < ym < (ny - 1.0):
        xycen = np.array((xm, ym), dtype=float)
    else:
        warnings.warn('quadratic polynomial maximum value falls outside '
                      'of the image', AstropyUserWarning)
        return np.array((np.nan, np.nan))

    return xycen


def centroid_sources(data, xpos, ypos, box_size=11, footprint=None, mask=None,
                     centroid_func=centroid_com, **kwargs):
    """
    Calculate the centroid of sources at the defined positions.

    A cutout image centered on each input position will be used to
    calculate the centroid position.  The cutout image is defined either
    using the ``box_size`` or ``footprint`` keyword.  The ``footprint``
    keyword can be used to create a non-rectangular cutout image.

    Parameters
    ----------
    data : 2D `~numpy.ndarray`
        The 2D image data. The image should be background-subtracted.

    xpos, ypos : float or array_like of float
        The initial ``x`` and ``y`` pixel position(s) of the center
        position.  A cutout image centered on this position be used to
        calculate the centroid.

    box_size : int or array_like of int, optional
        The size of the cutout image along each axis. If ``box_size`` is
        a number, then a square cutout of ``box_size`` will be created.
        If ``box_size`` has two elements, they must be in ``(ny, nx)``
        order. ``box_size`` must have odd values for both axes. Either
        ``box_size`` or ``footprint`` must be defined. If they are both
        defined, then ``footprint`` overrides ``box_size``.

    footprint : bool `~numpy.ndarray`, optional
        A 2D boolean array where `True` values describe the local
        footprint region to cutout. ``footprint`` can be used to create
        a non-rectangular cutout image, in which case the input ``xpos``
        and ``ypos`` represent the center of the minimal bounding box
        for the input ``footprint``. ``box_size=(n, m)`` is equivalent
        to ``footprint=np.ones((n, m))``. Either ``box_size`` or
        ``footprint`` must be defined. If they are both defined, then
        ``footprint`` overrides ``box_size``.

    mask : 2D bool `~numpy.ndarray`, optional
        A 2D boolean array with the same shape as ``data``, where a
        `True` value indicates the corresponding element of ``data`` is
        masked.

    error : 2D `~numpy.ndarray`, optional
        The 2D array of the 1-sigma errors of the input ``data``.
        ``error`` must have the same shape as ``data``.  ``error`` will
        be used only if supported by the input ``centroid_func``.

    centroid_func : callable, optional
        A callable object (e.g., function or class) that is used to
        calculate the centroid of a 2D array. The ``centroid_func``
        must accept a 2D `~numpy.ndarray`, have a ``mask`` keyword and
        optionally an ``error`` keyword. The callable object must return
        a tuple of two 1D `~numpy.ndarray`, representing the x and y
        centroids. The default is `~photutils.centroids.centroid_com`.

    **kwargs : `dict`
        Any additional keyword arguments accepted by the
        ``centroid_func``.

    Returns
    -------
    xcentroid, ycentroid : `~numpy.ndarray`
        The ``x`` and ``y`` pixel position(s) of the centroids. NaNs
        will be returned where the centroid failed. This is usually due
        a ``box_size`` that is too small when using a fitting-based
        centroid function (e.g., `centroid_1dg`, `centroid_2dg`, or
        `centroid_quadratic`.

    Examples
    --------
    >>> from photutils.centroids import centroid_com, centroid_sources
    >>> from photutils.datasets import make_4gaussians_image
    >>> from photutils.utils import circular_footprint
    >>> data = make_4gaussians_image()

    >>> x_init = (25, 91, 151, 160)
    >>> y_init = (40, 61, 24, 71)
    >>> footprint = circular_footprint(5.0)
    >>> x, y = centroid_sources(data, x_init, y_init, footprint=footprint,
    ...                         centroid_func=centroid_com)
    >>> print(x)  # doctest: +FLOAT_CMP
    [ 24.9865905   90.84751557 150.50228056 159.74319544]
    >>> print(y)  # doctest: +FLOAT_CMP
    [40.01096547 60.92509086 24.52986889 70.57971148]

    .. plot::
        :include-source:

        import matplotlib.pyplot as plt
        from photutils.centroids import centroid_com, centroid_sources
        from photutils.datasets import make_4gaussians_image
        from photutils.utils import circular_footprint
        data = make_4gaussians_image()
        x_init = (25, 91, 151, 160)
        y_init = (40, 61, 24, 71)
        footprint = circular_footprint(5.0)
        x, y = centroid_sources(data, x_init, y_init, footprint=footprint,
                                centroid_func=centroid_com)
        plt.figure(figsize=(8, 4))
        plt.imshow(data, origin='lower', interpolation='nearest')
        plt.scatter(x, y, marker='+', s=80, color='red', label='Centroids')
        plt.legend()
        plt.tight_layout()
    """
    xpos = np.atleast_1d(xpos)
    ypos = np.atleast_1d(ypos)
    if xpos.ndim != 1:
        raise ValueError('xpos must be a 1D array.')
    if ypos.ndim != 1:
        raise ValueError('ypos must be a 1D array.')

    if (np.any(np.min(xpos) < 0) or np.any(np.min(ypos) < 0)
            or np.any(np.max(xpos) > data.shape[1] - 1)
            or np.any(np.max(ypos) > data.shape[0] - 1)):
        raise ValueError('xpos, ypos values contains point(s) outside of '
                         'input data')

    if footprint is None:
        if box_size is None:
            raise ValueError('box_size or footprint must be defined.')
        box_size = as_pair('box_size', box_size, lower_bound=(0, 1),
                           check_odd=True)
        footprint = np.ones(box_size, dtype=bool)
    else:
        footprint = np.asanyarray(footprint, dtype=bool)
        if footprint.ndim != 2:
            raise ValueError('footprint must be a 2D array.')

    spec = inspect.signature(centroid_func)
    if 'mask' not in spec.parameters:
        raise ValueError('The input "centroid_func" must have a "mask" '
                         'keyword.')

    # drop any **kwargs not supported by the centroid_func
    centroid_kwargs = {}
    for key, val in kwargs.items():
        if key in spec.parameters:
            centroid_kwargs[key] = val

    xcentroids = []
    ycentroids = []
    for xp, yp in zip(xpos, ypos):
        slices_large, slices_small = overlap_slices(data.shape,
                                                    footprint.shape, (yp, xp))
        data_cutout = data[slices_large]

        footprint_mask = np.logical_not(footprint)
        # trim footprint mask if it has only partial overlap on the data
        footprint_mask = footprint_mask[slices_small]

        if mask is not None:
            # combine the input mask cutout and footprint mask
            mask_cutout = np.logical_or(mask[slices_large], footprint_mask)
        else:
            mask_cutout = footprint_mask

        if np.all(mask_cutout):
            raise ValueError(f'The cutout for the source at ({xp, yp}) is '
                             'completely masked. Please check your input '
                             'mask and footprint. Also note that footprint '
                             'must be a small, local footprint.')

        centroid_kwargs.update({'mask': mask_cutout})

        error = centroid_kwargs.get('error', None)
        if error is not None:
            centroid_kwargs['error'] = error[slices_large]

        # remove xpeak and ypeak from the dict and add back only if both
        # are specified and not None
        xpeak = centroid_kwargs.pop('xpeak', None)
        ypeak = centroid_kwargs.pop('ypeak', None)
        if xpeak is not None and ypeak is not None:
            centroid_kwargs['xpeak'] = xpeak - slices_large[1].start
            centroid_kwargs['ypeak'] = ypeak - slices_large[0].start

        try:
            xcen, ycen = centroid_func(data_cutout, **centroid_kwargs)
        except (ValueError, TypeError):
            xcen, ycen = np.nan, np.nan

        xcentroids.append(xcen + slices_large[1].start)
        ycentroids.append(ycen + slices_large[0].start)

    return np.array(xcentroids), np.array(ycentroids)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the core module.
"""

import astropy.units as u
import numpy as np
import pytest
from astropy.modeling.models import Gaussian2D
from astropy.utils.exceptions import AstropyUserWarning
from numpy.testing import assert_allclose, assert_equal

from photutils.aperture import CircularAperture
from photutils.profiles import CurveOfGrowth
from photutils.utils._optional_deps import HAS_MATPLOTLIB


@pytest.fixture(name='profile_data')
def fixture_profile_data():
    xsize = 101
    ysize = 80
    xcen = (xsize - 1) / 2
    ycen = (ysize - 1) / 2
    xycen = (xcen, ycen)

    sig = 10.0
    model = Gaussian2D(21., xcen, ycen, sig, sig)
    y, x = np.mgrid[0:ysize, 0:xsize]
    data = model(x, y)

    error = 10.0 * np.sqrt(data)
    mask = np.zeros(data.shape, dtype=bool)
    mask[:int(ycen), :int(xcen)] = True

    return xycen, data, error, mask


def test_curve_of_growth(profile_data):
    xycen, data, _, _ = profile_data

    radii = np.arange(1, 37)
    cg1 = CurveOfGrowth(data, xycen, radii, error=None, mask=None)

    assert_equal(cg1.radius, radii)
    assert cg1.area.shape == (36,)
    assert cg1.profile.shape == (36,)
    assert cg1.profile_error.shape == (0,)
    assert_allclose(cg1.area[0], np.pi)

    assert len(cg1.apertures) == 36
    assert isinstance(cg1.apertures[0], CircularAperture)

    radii = np.arange(1, 36)
    cg2 = CurveOfGrowth(data, xycen, radii, error=None, mask=None)
    assert cg2.area[0] > 0.0
    assert isinstance(cg2.apertures[0], CircularAperture)


def test_curve_of_growth_units(profile_data):
    xycen, data, error, _ = profile_data

    radii = np.arange(1, 36)
    unit = u.Jy
    cg1 = CurveOfGrowth(data << unit, xycen, radii, error=error << unit,
                        mask=None)

    assert cg1.profile.unit == unit
    assert cg1.profile_error.unit == unit

    with pytest.raises(ValueError):
        CurveOfGrowth(data << unit, xycen, radii, error=error, mask=None)


def test_curve_of_growth_error(profile_data):
    xycen, data, error, _ = profile_data

    radii = np.arange(1, 36)
    cg1 = CurveOfGrowth(data, xycen, radii, error=error, mask=None)

    assert cg1.profile.shape == (35,)
    assert cg1.profile_error.shape == (35,)


def test_curve_of_growth_mask(profile_data):
    xycen, data, error, mask = profile_data

    radii = np.arange(1, 36)
    cg1 = CurveOfGrowth(data, xycen, radii, error=error, mask=None)
    cg2 = CurveOfGrowth(data, xycen, radii, error=error, mask=mask)

    assert cg1.profile.sum() > cg2.profile.sum()
    assert np.sum(cg1.profile_error**2) > np.sum(cg2.profile_error**2)


def test_curve_of_growth_normalize(profile_data):
    xycen, data, _, _ = profile_data

    radii = np.arange(1, 36)
    cg1 = CurveOfGrowth(data, xycen, radii, error=None, mask=None)

    profile1 = cg1.profile
    cg1.normalize()
    profile2 = cg1.profile
    assert np.mean(profile2) < np.mean(profile1)

    cg1.normalize(method='sum')
    profile3 = cg1.profile
    assert np.mean(profile3) < np.mean(profile1)

    with pytest.raises(ValueError):
        cg1.normalize(method='invalid')

    cg1.__dict__['profile'] -= np.max(cg1.__dict__['profile'])
    msg = 'The profile cannot be normalized'
    with pytest.warns(AstropyUserWarning, match=msg):
        cg1.normalize(method='max')


def test_curve_of_growth_inputs(profile_data):
    xycen, data, error, _ = profile_data

    msg = 'radii must be > 0'
    with pytest.raises(ValueError, match=msg):
        radii = np.arange(10)
        CurveOfGrowth(data, xycen, radii, error=None, mask=None)

    msg = 'radii must be a 1D array and have at least two values'
    with pytest.raises(ValueError, match=msg):
        CurveOfGrowth(data, xycen, [1], error=None, mask=None)
    with pytest.raises(ValueError, match=msg):
        CurveOfGrowth(data, xycen, np.arange(1, 7).reshape(2, 3), error=None,
                      mask=None)

    msg = 'radii must be strictly increasing'
    with pytest.raises(ValueError, match=msg):
        radii = np.arange(1, 10)[::-1]
        CurveOfGrowth(data, xycen, radii, error=None, mask=None)

    with pytest.raises(ValueError):
        unit1 = u.Jy
        unit2 = u.km
        radii = np.arange(1, 36)
        CurveOfGrowth(data << unit1, xycen, radii, error=error << unit2)


@pytest.mark.skipif(not HAS_MATPLOTLIB, reason='matplotlib is required')
def test_curve_of_growth_plot(profile_data):
    xycen, data, error, _ = profile_data

    radii = np.arange(1, 36)
    cg1 = CurveOfGrowth(data, xycen, radii, error=None, mask=None)
    cg1.plot()
    with pytest.warns(AstropyUserWarning, match='Errors were not input'):
        cg1.plot_error()

    cg2 = CurveOfGrowth(data, xycen, radii, error=error, mask=None)
    cg2.plot()
    pc1 = cg2.plot_error()
    assert_allclose(pc1.get_facecolor(), [[0.5, 0.5, 0.5, 0.3]])
    pc2 = cg2.plot_error(facecolor='blue')
    assert_allclose(pc2.get_facecolor(), [[0, 0, 1, 1]])

    unit = u.Jy
    cg3 = CurveOfGrowth(data << unit, xycen, radii, error=error << unit,
                        mask=None)
    cg3.plot()
    cg3.plot_error()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This subpackage contains tools to perform point-spread-function (PSF)
photometry.
"""

from . import (epsf, epsf_stars, griddedpsfmodel, groupers, groupstars,
               models, photometry, photometry_depr, utils)
from .epsf import *  # noqa: F401, F403
from .epsf_stars import *  # noqa: F401, F403
from .griddedpsfmodel import *  # noqa: F401, F403
from .groupers import *  # noqa: F401, F403
from .groupstars import *  # noqa: F401, F403
from .matching import *  # noqa: F401, F403
from .models import *  # noqa: F401, F403
from .photometry import *  # noqa: F401, F403
from .photometry_depr import *  # noqa: F401, F403
from .utils import *  # noqa: F401, F403

# exclude matching from this list to avoid sphinx warnings
__all__ = []
__all__.extend(epsf.__all__)
__all__.extend(epsf_stars.__all__)
__all__.extend(griddedpsfmodel.__all__)
__all__.extend(groupers.__all__)
__all__.extend(groupstars.__all__)
__all__.extend(models.__all__)
__all__.extend(photometry.__all__)
__all__.extend(photometry_depr.__all__)
__all__.extend(utils.__all__)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This module provides tools for generating random (x, y) coordinates.
"""
import warnings
from collections import defaultdict

import numpy as np
from astropy.utils.exceptions import AstropyUserWarning


def apply_separation(xycoords, min_separation):
    from scipy.spatial import KDTree

    tree = KDTree(xycoords)
    pairs = tree.query_pairs(min_separation, output_type='ndarray')

    # create a dictionary of nearest neighbors (within min_separation)
    nn = {}
    nn = defaultdict(set)
    for p1, p2 in xycoords[pairs]:
        nn[tuple(p1)].add(tuple(p2))
        nn[tuple(p2)].add(tuple(p1))

    keep = []
    discard = set()
    for xycoord in xycoords:
        xycoord = tuple(xycoord)
        if xycoord not in discard:
            keep.append(xycoord)
            # remove nearest neighbors from the output
            neighbors = nn.get(xycoord, None)
            if neighbors is not None:
                discard.update(neighbors)

    return np.array(keep)


def make_random_xycoords(size, x_range, y_range, min_separation=0.0,
                         seed=None):
    """
    Make random (x, y) coordinates.

    Parameters
    ----------
    size : int
        The number of coordinates to generate.

    x_range : tuple
        The range of x values (min, max).

    y_range : tuple
        The range of y values (min, max).

    min_separation : float, optional
        The minimum separation in pixels between coordinates.

    seed : int, optional
        A seed to initialize the `numpy.random.BitGenerator`. If `None`,
        then fresh, unpredictable entropy will be pulled from the OS.

    Returns
    -------
    xycoords : `~numpy.ndarray`
        The (x, y) random coordinates with shape ``(size, 2)``.
    """
    if min_separation > 0:
        # scale the number of random coordinates to account for
        # some being discarded due to min_separation
        ncoords = size * 10

    rng = np.random.default_rng(seed)
    xc = rng.uniform(x_range[0], x_range[1], ncoords)
    yc = rng.uniform(y_range[0], y_range[1], ncoords)
    xycoords = np.transpose(np.array((xc, yc)))

    xycoords = apply_separation(xycoords, min_separation)
    xycoords = xycoords[:size]

    if len(xycoords) < size:
        warnings.warn(f'Unable to produce {size!r} coordinates.',
                      AstropyUserWarning)

    return xycoords



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This module provides tools for detecting sources in an image.
"""

import warnings

import numpy as np
from astropy.stats import SigmaClip

from photutils.segmentation.core import SegmentationImage
from photutils.segmentation.utils import _make_binary_structure
from photutils.utils._quantity_helpers import process_quantities
from photutils.utils._stats import nanmean, nanstd
from photutils.utils.exceptions import NoDetectionsWarning

__all__ = ['detect_threshold', 'detect_sources']


def detect_threshold(data, nsigma, *, background=None, error=None, mask=None,
                     sigma_clip=SigmaClip(sigma=3.0, maxiters=10)):
    """
    Calculate a pixel-wise threshold image that can be used to detect
    sources.

    This is a simple convenience function that uses sigma-clipped
    statistics to compute a scalar background and noise estimate. In
    general, one should perform more sophisticated estimates, e.g.,
    using `~photutils.background.Background2D`.

    Parameters
    ----------
    data : 2D `~numpy.ndarray`
        The 2D array of the image.

    nsigma : float
        The number of standard deviations per pixel above the
        ``background`` for which to consider a pixel as possibly being
        part of a source.

    background : float or 2D `~numpy.ndarray`, optional
        The background value(s) of the input ``data``. ``background``
        may either be a scalar value or a 2D array with the same
        shape as the input ``data``. If the input ``data`` has been
        background-subtracted, then set ``background`` to ``0.0`` (this
        should be typical). If `None`, then a scalar background value
        will be estimated as the sigma-clipped image mean.

    error : float or 2D `~numpy.ndarray`, optional
        The Gaussian 1-sigma standard deviation of the background
        noise in ``data``. ``error`` should include all sources of
        "background" error, but *exclude* the Poisson error of the
        sources. If ``error`` is a 2D image, then it should represent
        the 1-sigma background error in each pixel of ``data``. If
        `None`, then a scalar background rms value will be estimated
        as the sigma-clipped image standard deviation.

    mask : 2D bool `~numpy.ndarray`, optional
        A boolean mask with the same shape as ``data``, where a `True`
        value indicates the corresponding element of ``data`` is masked.
        Masked pixels are ignored when computing the image background
        statistics.

    sigma_clip : `astropy.stats.SigmaClip` instance, optional
        A `~astropy.stats.SigmaClip` object that defines the sigma
        clipping parameters.

    Returns
    -------
    threshold : 2D `~numpy.ndarray`
        A 2D image with the same shape as ``data`` containing the
        pixel-wise threshold values.

    See Also
    --------
    :class:`photutils.background.Background2D`
    :func:`photutils.segmentation.detect_sources`
    :class:`photutils.segmentation.SourceFinder`

    Notes
    -----
    The ``mask`` and ``sigma_clip`` inputs are used only if it
    is necessary to estimate ``background`` or ``error`` using
    sigma-clipped background statistics. If ``background`` and ``error``
    are both input, then ``mask`` and ``sigma_clip`` are ignored.
    """
    arrays, unit = process_quantities((data, background, error),
                                      ('data', 'background', 'error'))
    data, background, error = arrays

    if not isinstance(sigma_clip, SigmaClip):
        raise TypeError('sigma_clip must be a SigmaClip object')

    if background is None or error is None:
        if mask is not None:
            data = np.ma.MaskedArray(data, mask)

        clipped_data = sigma_clip(data, masked=False, return_bounds=False,
                                  copy=True)

    if background is None:
        background = nanmean(clipped_data)

    if not np.isscalar(background) and background.shape != data.shape:
        raise ValueError('If input background is 2D, then it must have the '
                         'same shape as the input data.')

    if error is None:
        error = nanstd(clipped_data)
    if not np.isscalar(error) and error.shape != data.shape:
        raise ValueError('If input error is 2D, then it must have the same '
                         'shape as the input data.')

    threshold = (np.broadcast_to(background, data.shape)
                 + np.broadcast_to(error * nsigma, data.shape))

    if unit:
        threshold <<= unit

    return threshold


def _detect_sources(data, thresholds, npixels, footprint, inverse_mask, *,
                    deblend_mode=False):
    """
    Detect sources above a specified threshold value in an image.

    Detected sources must have ``npixels`` connected pixels that are
    each greater than the ``threshold`` value.  If the filtering option
    is used, then the ``threshold`` is applied to the filtered image.
    The input ``mask`` can be used to mask pixels in the input data.
    Masked pixels will not be included in any source.

    This function does not deblend overlapping sources.  First use this
    function to detect sources followed by
    :func:`~photutils.segmentation.deblend_sources` to deblend sources.

    Parameters
    ----------
    data : 2D `~numpy.ndarray`
        The 2D array of the image. If filtering is desired, please input
        a convolved image here.

    thresholds : list of 2D `~numpy.ndarray` or 1D array of floats
        The data values (as a 1D array of floats) or pixel-wise data
        values (as a sequence of 2D arrays) to be used for the detection
        thresholds. 2D threshold arrays must have the same shape as
        ``data``.

    npixels : int
        The minimum number of connected pixels, each greater than
        ``threshold``, that an object must have to be detected.
        ``npixels`` must be a positive integer.

    footprint : array_like
        A footprint that defines feature connections. As an example,
        for connectivity along pixel edges only, the footprint is
        ``np.array([[0, 1, 0]], [1, 1, 1], [0, 1, 0]])``.

    inverse_mask : 2D bool `~numpy.ndarray`
        A boolean mask, with the same shape as the input ``data``, where
        `False` values indicate masked pixels (the inverse of usual
        pixel masks). Masked pixels will not be included in any source.

    deblend_mode : bool, optional
        If `True` do not include the segmentation image in the output
        list for any threshold level where the number of detected
        sources is less than 2. The deblend mode also does not relabel
        the output segmentation image to have consecutive label. This
        keyword improves performance of source deblending.

    Returns
    -------
    segment_image : list of `~photutils.segmentation.SegmentationImage`
        A list of 2D segmentation images, one for each input threshold,
        with the same shape as ``data``, where sources are marked
        by different positive integer values. A value of zero is
        reserved for the background. If no sources are found for a given
        threshold, then the output list will contain `None` for that
        threshold. Also see the ``deblend_mode`` keyword.
    """
    from scipy.ndimage import find_objects
    from scipy.ndimage import label as ndi_label

    segms = []
    for threshold in thresholds:
        # RuntimeWarning caused by > comparison when data contains NaNs
        # is ignored when calling _detect_sources
        segment_img = data > threshold

        if inverse_mask is not None:
            segment_img &= inverse_mask

        # return if threshold was too high to detect any sources
        if np.count_nonzero(segment_img) == 0:
            if not deblend_mode:
                warnings.warn('No sources were found.', NoDetectionsWarning)
                segms.append(None)
            continue

        # recasting segment_img to int and using output=segment_img
        # gives similar performance
        segment_img, nlabels = ndi_label(segment_img, structure=footprint)
        labels = np.arange(nlabels) + 1

        # remove objects with less than npixels
        # NOTE: making cutout images and setting their pixels to 0 is
        # ~10x faster than using segment_img directly and ~50% faster
        # than using ndimage.sum_labels.
        slices = find_objects(segment_img)
        segm_labels = []
        segm_slices = []
        for label, slc in zip(labels, slices):
            cutout = segment_img[slc]
            segment_mask = (cutout == label)
            if np.count_nonzero(segment_mask) < npixels:
                cutout[segment_mask] = 0
                continue
            segm_labels.append(label)
            segm_slices.append(slc)

        if np.count_nonzero(segment_img) == 0:
            if not deblend_mode:
                warnings.warn('No sources were found.', NoDetectionsWarning)
                segms.append(None)
            continue

        if not deblend_mode:
            # relabel the segmentation image with consecutive numbers
            nlabels = len(segm_labels)
            if len(labels) != nlabels:
                label_map = np.zeros(np.max(labels) + 1, dtype=int)
                labels = np.arange(nlabels) + 1
                label_map[segm_labels] = labels
                segment_img = label_map[segment_img]
        else:
            labels = segm_labels

        segm = object.__new__(SegmentationImage)
        segm._data = segment_img
        segm.__dict__['labels'] = labels
        segm.__dict__['slices'] = segm_slices

        if deblend_mode and segm.nlabels == 1:
            continue

        segms.append(segm)

    return segms


def detect_sources(data, threshold, npixels, *, connectivity=8, mask=None):
    """
    Detect sources above a specified threshold value in an image.

    Detected sources must have ``npixels`` connected pixels that are
    each greater than the ``threshold`` value.  If the filtering option
    is used, then the ``threshold`` is applied to the filtered image.
    The input ``mask`` can be used to mask pixels in the input data.
    Masked pixels will not be included in any source.

    This function does not deblend overlapping sources.  First use this
    function to detect sources followed by
    :func:`~photutils.segmentation.deblend_sources` to deblend sources.

    Parameters
    ----------
    data : 2D `~numpy.ndarray`
        The 2D array of the image. If filtering is desired, please input
        a convolved image here.

    threshold : float or 2D `~numpy.ndarray`
        The data value or pixel-wise data values to be used for the
        detection threshold. A 2D ``threshold`` array must have the same
        shape and units as ``data``.

    npixels : int
        The minimum number of connected pixels, each greater than
        ``threshold``, that an object must have to be detected.
        ``npixels`` must be a positive integer.

    connectivity : {4, 8}, optional
        The type of pixel connectivity used in determining how pixels
        are grouped into a detected source. The options are 4 or
        8 (default). 4-connected pixels touch along their edges.
        8-connected pixels touch along their edges or corners.

    mask : 2D bool `~numpy.ndarray`, optional
        A boolean mask, with the same shape as the input ``data``, where
        `True` values indicate masked pixels. Masked pixels will not be
        included in any source.

    Returns
    -------
    segment_image : `~photutils.segmentation.SegmentationImage` or `None`
        A 2D segmentation image, with the same shape as ``data``, where
        sources are marked by different positive integer values. A value
        of zero is reserved for the background. If no sources are found
        then `None` is returned.

    See Also
    --------
    :func:`photutils.segmentation.deblend_sources`
    :class:`photutils.segmentation.SourceFinder`

    Examples
    --------
    .. plot::
        :include-source:

        import matplotlib.pyplot as plt
        from astropy.convolution import convolve
        from astropy.stats import sigma_clipped_stats
        from astropy.visualization import simple_norm
        from photutils.datasets import make_100gaussians_image
        from photutils.segmentation import (detect_sources,
                                            make_2dgaussian_kernel)

        # make a simulated image
        data = make_100gaussians_image()

        # use sigma-clipped statistics to (roughly) estimate the background
        # background noise levels
        mean, _, std = sigma_clipped_stats(data)

        # subtract the background
        data -= mean

        # detect the sources
        threshold = 3. * std
        kernel = make_2dgaussian_kernel(3.0, size=3)  # FWHM = 3.
        convolved_data = convolve(data, kernel)
        segm = detect_sources(convolved_data, threshold, npixels=5)

        # plot the image and the segmentation image
        fig, (ax1, ax2) = plt.subplots(2, 1, figsize=(8, 10))
        norm = simple_norm(data, 'sqrt', percent=99.)
        ax1.imshow(data, origin='lower', interpolation='nearest',
                   norm=norm)
        ax2.imshow(segm.data, origin='lower', interpolation='nearest',
                   cmap=segm.make_cmap(seed=1234))
        plt.tight_layout()
    """
    _ = process_quantities((data, threshold), ('data', 'threshold'))

    if (npixels <= 0) or (int(npixels) != npixels):
        raise ValueError('npixels must be a positive integer, got '
                         f'"{npixels}"')

    if mask is not None:
        if mask.shape != data.shape:
            raise ValueError('mask must have the same shape as the input '
                             'image.')
        if mask.all():
            raise ValueError('mask must not be True for every pixel. There '
                             'are no unmasked pixels in the image to detect '
                             'sources.')
        inverse_mask = np.logical_not(mask)
    else:
        inverse_mask = None

    footprint = _make_binary_structure(data.ndim, connectivity)

    with warnings.catch_warnings():
        warnings.simplefilter('ignore', category=RuntimeWarning)
        return _detect_sources(data, (threshold,), npixels, footprint,
                               inverse_mask, deblend_mode=False)[0]



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This module provides Quantity helper tools.
"""


def process_quantities(arrays, names):
    """
    Check units of input arrays.

    If any of the input arrays have units then they all must have
    units and the units must be the same.

    Unitless `~numpy.ndarray` objects are returned along with the array
    unit.

    Parameters
    ----------
    arrays : list of `~numpy.ndarray` or `~astropy.units.Quantity`
        A list of arrays.

    names : list of str
        A list of names corresponding to the input ``arrays``.

    Returns
    -------
    arrays : list of `~numpy.ndarray`
        A list of numpy arrays, where units have been removed.

    unit : `~astropy.unit.Unit`
        The data unit. `None` will be returned if the input arrays do
        not have units.

    Raises
    ------
    ValueError
        If the input arrays do not all have the same units.
    """
    unit = {getattr(arr, 'unit', None) for arr in arrays if arr is not None}
    if len(unit) > 1:
        if len(names) == 2:
            str_names = f'{names[0]} or {names[1]}'
        elif len(names) > 2:
            names1 = ', '.join(names[:-1])
            str_names = f'{names1}, or {names[-1]}'
        raise ValueError(f'If {str_names} has units, then they must have the '
                         'same units.')
    unit = unit.pop()
    if unit is not None:
        arrays = [arr.value if arr is not None else arr for arr in arrays]
    return arrays, unit
